"""Off-policy Actor-Critic with Normalizing Flows density approximation."""
from raylab.utils.dictionaries import deep_merge

from .abstract import AbstractActorCritic
from .mixins import ActionValueMixin
from .mixins import MaximumEntropyMixin
from .mixins import NormalizingFlowActorMixin


BASE_CONFIG = {"actor": {}, "critic": {}, "entropy": {}}


class OffPolicyNFAC(
    NormalizingFlowActorMixin,
    ActionValueMixin,
    MaximumEntropyMixin,
    AbstractActorCritic,
):
    """Actor-Critic module with stochastic actor and action-value critics."""

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config, False, ["actor", "critic", "entropy"])
        super().__init__(obs_space, action_space, config)



# pylint:disable=missing-docstring
# pylint:enable=missing-docstring
import math
from typing import Dict
from typing import List

import torch
import torch.nn as nn
from ray.rllib.utils import override

from .abstract import ConditionalDistribution


class Normal(ConditionalDistribution):
    """
    Creates a normal (also called Gaussian) distribution parameterized by
    :attr:`loc` and :attr:`scale`.
    """

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        loc, scale = torch.chunk(inputs, 2, dim=-1)
        return {"loc": loc, "scale": scale}

    @override(ConditionalDistribution)
    @torch.jit.export
    def sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        out = self._gen_sample(params, sample_shape).detach()
        return out, self.log_prob(out, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def rsample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        out = self._gen_sample(params, sample_shape)
        return out, self.log_prob(out, params)

    def _gen_sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int]):
        loc, scale = self._unpack_params(params)
        shape = sample_shape + loc.shape
        eps = torch.randn(shape, dtype=loc.dtype, device=loc.device)
        return loc + eps * scale

    @override(ConditionalDistribution)
    @torch.jit.export
    def log_prob(self, value, params: Dict[str, torch.Tensor]):
        loc, scale = self._unpack_params(params)
        var = scale ** 2
        const = math.log(math.sqrt(2 * math.pi))
        return -((value - loc) ** 2) / (2 * var) - scale.log() - const

    @override(ConditionalDistribution)
    @torch.jit.export
    def cdf(self, value, params: Dict[str, torch.Tensor]):
        loc, scale = self._unpack_params(params)
        return 0.5 * (1 + torch.erf((value - loc) * scale.reciprocal() / math.sqrt(2)))

    @override(ConditionalDistribution)
    @torch.jit.export
    def icdf(self, value, params: Dict[str, torch.Tensor]):
        loc, scale = self._unpack_params(params)
        return loc + scale * torch.erfinv(2 * value - 1) * math.sqrt(2)

    @override(ConditionalDistribution)
    @torch.jit.export
    def entropy(self, params: Dict[str, torch.Tensor]):
        _, scale = self._unpack_params(params)
        return 0.5 + 0.5 * math.log(2 * math.pi) + torch.log(scale)

    @override(ConditionalDistribution)
    @torch.jit.export
    def reproduce(self, value, params: Dict[str, torch.Tensor]):
        loc, scale = self._unpack_params(params)
        eps = (value - loc) / scale
        sample_ = loc + scale * eps.detach()
        return sample_, self.log_prob(sample_, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def deterministic(self, params: Dict[str, torch.Tensor]):
        loc, _ = self._unpack_params(params)
        sample = loc
        return sample, self.log_prob(sample, params)

    def _unpack_params(self, params: Dict[str, torch.Tensor]):
        # pylint:disable=no-self-use
        return params["loc"], params["scale"]



"""Losses for Importance Sampled Fitted V Iteration."""
from typing import Tuple

import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from ray.rllib.utils import override
from torch import Tensor

import raylab.utils.dictionaries as dutil
from raylab.policy.modules.actor.policy.stochastic import StochasticPolicy
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

from .abstract import Loss


class ISFittedVIteration(Loss):
    """Loss function for Importance Sampled Fitted V Iteration.

    Args:
        critic: state-value function
        target_critic: state-value function for the next state

    Attributes:
        gamma: discount factor
    """

    IS_RATIOS = "is_ratios"
    batch_keys: Tuple[str, str] = (SampleBatch.CUR_OBS, "is_ratios")
    gamma: float = 0.99

    def __init__(self, critic: nn.Module, target_critic: nn.Module):
        self.critic = critic
        self.target_critic = target_critic

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute loss for importance sampled fitted V iteration."""
        obs, is_ratios = dutil.get_keys(batch, *self.batch_keys)

        values = self.critic(obs).squeeze(-1)
        with torch.no_grad():
            targets = self.sampled_one_step_state_values(batch)
        value_loss = torch.mean(
            is_ratios * torch.nn.MSELoss(reduction="none")(values, targets) / 2
        )
        return value_loss, {"loss(critic)": value_loss.item()}

    def sampled_one_step_state_values(self, batch: TensorDict) -> Tensor:
        """Bootstrapped approximation of true state-value using sampled transition."""
        next_obs, rewards, dones = dutil.get_keys(
            batch, SampleBatch.NEXT_OBS, SampleBatch.REWARDS, SampleBatch.DONES,
        )
        return torch.where(
            dones,
            rewards,
            rewards + self.gamma * self.target_critic(next_obs).squeeze(-1),
        )


class ISSoftVIteration(ISFittedVIteration):
    """Loss function for Importance Sampled Soft V Iteration.

    Args:
        critic: state-value function
        target_critic: state-value function for the next state
        actor: stochastic policy

    Attributes:
        gamma: discount factor
        alpha: entropy coefficient
    """

    # pylint:disable=too-few-public-methods
    ENTROPY = "entropy"
    gamma: float = 0.99
    alpha: float = 0.05

    def __init__(
        self, critic: nn.Module, target_critic: nn.Module, actor: StochasticPolicy
    ):
        super().__init__(critic, target_critic)
        self.actor = actor

    @override(ISFittedVIteration)
    def sampled_one_step_state_values(self, batch: TensorDict) -> Tensor:
        """Bootstrapped approximation of true state-value using sampled transition."""
        if self.ENTROPY in batch:
            entropy = batch[self.ENTROPY]
        else:
            with torch.no_grad():
                _, logp = self.actor(batch[SampleBatch.CUR_OBS])
                entropy = -logp

        next_obs, rewards, dones = dutil.get_keys(
            batch, SampleBatch.NEXT_OBS, SampleBatch.REWARDS, SampleBatch.DONES,
        )
        augmented_rewards = rewards + self.alpha * entropy
        return torch.where(
            dones,
            augmented_rewards,
            augmented_rewards + self.gamma * self.target_critic(next_obs).squeeze(-1),
        )



# pylint:disable=missing-module-docstring
import functools
from typing import Callable
from typing import Union

import torch.nn as nn

from . import activation as activations


def get_activation(activation: Union[str, dict, None]) -> Callable[[], nn.Module]:
    """Return activation module constructor from specification.

    Args:
        activation: the activation function's specification. Can be either
            a string with the activation module class' name, a dict with the
            name as the `name` field and additional keyword arguments, or None,
            in which case the indentity module class is returned.

    Raises:
        ValueError: If the type corresponding to the activation's name cannot
            be found.
    """
    if activation is None:
        return nn.Identity

    if isinstance(activation, dict):
        name = activation["name"]
        options = activation.get("options", {})
    else:
        name = activation
        options = {}

    if name in dir(nn.modules.activation):
        cls = getattr(nn.modules.activation, name)
        if issubclass(cls, nn.Module):
            return functools.partial(cls, **options)

    if name in dir(activations):
        cls = getattr(activations, name)
        if issubclass(cls, nn.Module):
            return functools.partial(cls, **options)

    raise ValueError(f"Couldn't find activation with name '{name}'")



"""
Soft Actor-Critic: Off-Policy Maximum Entropy Deep Reinforcement Learning
with a Stochastic Actor.
"""
from raylab.agents import trainer
from raylab.agents.off_policy import OffPolicyTrainer

from .policy import SACTorchPolicy


def sac_config(cls: type) -> type:
    """Add configurations for Soft Actor-Critic-based agents."""

    for config_setter in [
        trainer.option(
            "target_entropy",
            None,
            help="""Target entropy for temperature parameter optimization.

            If 'auto', will use the heuristic provided in the SAC paper,
            H = -dim(A), where A is the action space""",
        ),
        trainer.option("torch_optimizer/actor", {"type": "Adam", "lr": 1e-3}),
        trainer.option("torch_optimizer/critics", {"type": "Adam", "lr": 1e-3}),
        trainer.option("torch_optimizer/alpha", {"type": "Adam", "lr": 1e-3}),
        trainer.option(
            "polyak",
            0.995,
            help="Interpolation factor in polyak averaging for target networks.",
        ),
        trainer.option(
            "exploration_config/type",
            "raylab.utils.exploration.StochasticActor",
            override=True,
        ),
    ]:
        cls = config_setter(cls)

    return cls


@trainer.configure
@sac_config
@trainer.option("module", {"type": "SAC", "critic": {"double_q": True}}, override=True)
@trainer.option("exploration_config/pure_exploration_steps", 1000)
@trainer.option("evaluation_config/explore", False, override=True)
class SACTrainer(OffPolicyTrainer):
    """Single agent trainer for SAC."""

    _name = "SoftAC"
    _policy = SACTorchPolicy



"""SVG(inf) policy class using PyTorch."""
from contextlib import contextmanager

import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from ray.rllib.utils import override

from raylab.agents.svg import SVGTorchPolicy
from raylab.policy import AdaptiveKLCoeffMixin
from raylab.policy import EnvFnMixin
from raylab.policy.losses import TrajectorySVG
from raylab.pytorch.optim import build_optimizer


class SVGInfTorchPolicy(AdaptiveKLCoeffMixin, SVGTorchPolicy):
    """Stochastic Value Gradients policy for full trajectories."""

    # pylint:disable=abstract-method

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        # Flag for off-policy learning
        self._off_policy_learning = False

        self.loss_actor = TrajectorySVG(
            self.module.model, self.module.actor, self.module.critic,
        )

    @property
    @override(SVGTorchPolicy)
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.svg.inf import SVGInfTrainer

        return SVGInfTrainer.options

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(SVGTorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["torch_optimizer"]
        component_map = {
            "on_policy": self.module.actor,
            "off_policy": nn.ModuleList([self.module.model, self.module.critic]),
        }

        mapping = {
            name: build_optimizer(module, config[name])
            for name, module in component_map.items()
        }

        optimizers.update(mapping)
        return optimizers

    @override(SVGTorchPolicy)
    def compile(self):
        super().compile()
        self.loss_actor.compile()

    @override(SVGTorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        if self._off_policy_learning:
            info = self._learn_off_policy(batch_tensors)
        else:
            info = self._learn_on_policy(batch_tensors, samples)
        info.update(self.extra_grad_info(batch_tensors))
        return info

    @contextmanager
    def learning_off_policy(self):
        """Signal to policy to use samples for updating off-policy components."""
        old = self._off_policy_learning
        self._off_policy_learning = True
        yield
        self._off_policy_learning = old

    def _learn_off_policy(self, batch_tensors):
        """Update off-policy components."""
        batch_tensors, info = self.add_truncated_importance_sampling_ratios(
            batch_tensors
        )

        with self.optimizers.optimize("off_policy"):
            loss, _info = self.compute_joint_model_value_loss(batch_tensors)
            info.update(_info)
            loss.backward()

        self._update_polyak()
        return info

    def _learn_on_policy(self, batch_tensors, samples):
        """Update on-policy components."""
        episodes = [self.lazy_tensor_dict(s) for s in samples.split_by_episode()]

        with self.optimizers.optimize("on_policy"):
            loss, info = self.loss_actor(episodes)
            kl_div = self._avg_kl_divergence(batch_tensors)
            loss = loss + kl_div * self.curr_kl_coeff
            loss.backward()

        info.update(self.update_kl_coeff(samples))
        return info

    @torch.no_grad()
    @override(AdaptiveKLCoeffMixin)
    def _kl_divergence(self, sample_batch):
        batch_tensors = self.lazy_tensor_dict(sample_batch)
        return self._avg_kl_divergence(batch_tensors).item()

    def _avg_kl_divergence(self, batch_tensors):
        logp = self.module.actor.log_prob(
            batch_tensors[SampleBatch.CUR_OBS], batch_tensors[SampleBatch.ACTIONS]
        )
        return torch.mean(batch_tensors[SampleBatch.ACTION_LOGP] - logp)

    @torch.no_grad()
    def extra_grad_info(self, batch_tensors):
        """Compute gradient norm for components. Also clips on-policy gradient."""
        if self._off_policy_learning:
            model_params = self.module.model.parameters()
            value_params = self.module.critic.parameters()
            fetches = {
                "model_grad_norm": nn.utils.clip_grad_norm_(
                    model_params, float("inf")
                ).item(),
                "value_grad_norm": nn.utils.clip_grad_norm_(
                    value_params, float("inf")
                ).item(),
            }
        else:
            policy_params = self.module.actor.parameters()
            fetches = {
                "policy_grad_norm": nn.utils.clip_grad_norm_(
                    policy_params, max_norm=self.config["max_grad_norm"]
                ).item(),
                "policy_entropy": -batch_tensors[SampleBatch.ACTION_LOGP].mean().item(),
                "curr_kl_coeff": self.curr_kl_coeff,
            }
        return fetches



"""Stochastic Model with Normalizing Flows density approximation."""
import warnings

import gym.spaces as spaces
import torch.nn as nn
from ray.rllib.utils import override

import raylab.policy.modules.networks as networks
import raylab.pytorch.nn as nnx
import raylab.pytorch.nn.distributions as ptd
from raylab.utils.dictionaries import deep_merge

from .stochastic_model_mixin import StochasticModel
from .stochastic_model_mixin import StochasticModelMixin


BASE_CONFIG = {
    # Number of independent models. If 0 return a single model accessible via the
    # 'model' atribute. If greater than 0, return a nn.ModuleList of models accessible
    # via the 'models' attribute.
    "ensemble_size": 0,
    "residual": True,
    "conditional_prior": True,
    "input_encoder": {"units": (64, 64), "activation": "ReLU"},
    "num_flows": 4,
    "conditional_flow": False,
    "flow": {
        "type": "AffineCouplingTransform",
        "transform_net": {"type": "MLP", "num_blocks": 0},
    },
}


class NormalizingFlowModelMixin(StochasticModelMixin):
    """Stochastic model module with Normalizing Flow density estimator."""

    # pylint:disable=too-few-public-methods

    @override(StochasticModelMixin)
    def _make_model(self, obs_space, action_space, config):
        assert isinstance(
            obs_space, spaces.Box
        ), f"Normalizing Flow incompatible with observation space {type(obs_space)}"
        return super()._make_model(obs_space, action_space, config)

    @staticmethod
    @override(StochasticModelMixin)
    def process_config(config):
        return deep_merge(
            BASE_CONFIG,
            config.get("model", {}),
            False,
            ["input_encoder", "flow", "initializer_options"],
            ["flow"],
        )

    @override(StochasticModelMixin)
    def build_single_model(self, obs_space, action_space, config):
        # PRIOR ========================================================================
        params_module, base_dist = self._make_model_prior(
            obs_space, action_space, config
        )
        # NormalizingFlow ==============================================================
        transforms = self._make_model_transforms(
            obs_space, params_module.state_size, config
        )
        dist_module = ptd.TransformedDistribution(
            base_dist=base_dist, transform=ptd.flows.CompositeTransform(transforms),
        )

        return StochasticModel.assemble(params_module, dist_module, config)

    @staticmethod
    def _make_model_prior(obs_space, action_space, config):
        # Ensure we're not encoding the inputs for nothing
        obs_size, act_size = obs_space.shape[0], action_space.shape[0]
        if config["conditional_prior"] or config["conditional_flow"]:
            input_encoder = nnx.StateActionEncoder(
                obs_size, act_size, **config["input_encoder"]
            )
        else:
            warnings.warn("Model is blind to the observations")
            input_encoder = nnx.StateActionEncoder(obs_size, act_size, units=())

        params_module = NFNormalParams(input_encoder, obs_space, config)
        base_dist = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)
        return params_module, base_dist

    @staticmethod
    def _make_model_transforms(obs_space, state_size, config):
        obs_size = obs_space.shape[0]
        flow_config = config["flow"].copy()
        cls = getattr(ptd.flows, flow_config.pop("type"))

        if issubclass(cls, ptd.flows.CouplingTransform):
            net_config = flow_config.pop("transform_net")
            net_config.setdefault("hidden_features", obs_size)
            transform_net = getattr(networks, net_config.pop("type"))

            def transform_net_create_fn(in_features, out_features):
                return transform_net(
                    in_features,
                    out_features,
                    state_features=state_size if config["conditional_flow"] else None,
                    **net_config,
                )

            masks = [
                ptd.flows.masks.create_alternating_binary_mask(obs_size, bool(i % 2))
                for i in range(config["num_flows"])
            ]
            transforms = [cls(m, transform_net_create_fn, **flow_config) for m in masks]

        else:
            raise NotImplementedError(f"Unsupported flow type {cls}")

        return transforms


class NFNormalParams(nn.Module):
    """Maps inputs to distribution parameters for Normalizing Flows."""

    def __init__(self, input_encoder, obs_space, config):
        super().__init__()
        self.input_encoder = input_encoder
        self.state_size = self.input_encoder.out_features

        obs_size = obs_space.shape[0]
        if config["conditional_prior"]:
            self.params = nnx.NormalParams(self.state_size, obs_size)
        else:
            self.params = nnx.StdNormalParams(1, obs_size)

    @override(nn.Module)
    def forward(self, obs, act):  # pylint:disable=arguments-differ
        state = self.input_encoder(obs, act)
        params = self.params(state)
        # For use in conditional flows later
        params["state"] = state
        return params



"""Losses for Policy-Aware Model Learning."""
from typing import Tuple

import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from torch import Tensor
from torch.autograd import grad

from raylab.policy.modules.actor.policy.stochastic import StochasticPolicy
from raylab.policy.modules.critic.q_value import QValueEnsemble
from raylab.policy.modules.model.stochastic.ensemble import StochasticModelEnsemble
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin
from .mle import ModelEnsembleMLE


class SPAML(EnvFunctionsMixin, Loss):
    """Soft Policy-iteration-Aware Model Learning.

    Computes the decision-aware loss for model ensembles used in Model-Aware
    Policy Optimization.

    Args:
        models: The stochastic model ensemble
        actor: The stochastic policy
        critics: The action-value estimators

    Attributes:
        gamma: Discount factor
        alpha: Entropy regularization coefficient
        grad_estimator: Gradient estimator for expecations ('PD' or 'SF')
        lambda_: Kullback Leibler regularization coefficient

    Note:
        `N` denotes the size of the model ensemble, `O` the size of the
        observation, and `A` the size of the action
    """

    batch_keys: Tuple[str, str, str] = (
        SampleBatch.CUR_OBS,
        SampleBatch.ACTIONS,
        SampleBatch.NEXT_OBS,
    )
    gamma: float = 0.99
    alpha: float = 0.05
    grad_estimator: str = "SF"
    lambda_: float = 0.05

    def __init__(
        self,
        models: StochasticModelEnsemble,
        actor: StochasticPolicy,
        critics: QValueEnsemble,
    ):
        super().__init__()
        modules = nn.ModuleDict()
        modules["models"] = models
        modules["policy"] = actor
        modules["critics"] = critics
        self._modules = modules
        self._loss_mle = ModelEnsembleMLE(models)

    @property
    def initialized(self) -> bool:
        """Whether or not the loss setup is complete."""
        return self._env.initialized

    @property
    def ensemble_size(self) -> int:
        """The number of models in the ensemble."""
        return len(self._modules["models"])

    def compile(self):
        self._modules.update({k: torch.jit.script(m) for k, m in self._modules.items()})
        self._loss_mle.compile()

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        assert self.initialized, (
            "Environment functions missing. "
            "Did you set reward and termination functions?"
        )
        obs = batch[SampleBatch.CUR_OBS]
        obs = self.expand_foreach_model(obs)
        action = self.generate_action(obs)
        value_target = self.zero_step_action_value(obs, action)
        value_pred = self.one_step_action_value_surrogate(obs, action)
        grad_loss = self.action_gradient_loss(action, value_target, value_pred)
        mle_loss = self.maximum_likelihood_loss(batch)

        loss = grad_loss + self.lambda_ * mle_loss
        info = {f"loss(models[{i}])": s for i, s in enumerate(loss.tolist())}
        info["loss(daml)"] = grad_loss.mean().item()
        info["loss(mle)"] = mle_loss.mean().item()
        return loss, info

    def expand_foreach_model(self, tensor: Tensor) -> Tensor:
        """Add first dimension to tensor with the size of the model ensemble.

        Args:
            tensor: Tensor of shape `S`

        Returns:
            Tensor `tensor` expanded to shape `(N,) + S`
        """
        return tensor.expand((len(self._modules["models"]),) + tensor.shape)

    @torch.no_grad()
    def generate_action(self, obs: Tensor) -> Tensor:
        """Given state, sample action with the stochastic policy.

        Generates one action for each of the `N` models in the ensemble, so that
        action gradients and subsequent losses may be calculated for each model.

        Args:
            obs: The current observation tensor of shape `(N, *, O)`

        Returns:
            The action tensor of shape `(N, *, A)` with `requires_grad` enabled
        """
        action, _ = self._modules["policy"].sample(obs)
        return action.requires_grad_(True)

    def zero_step_action_value(self, obs: Tensor, action: Tensor) -> Tensor:
        """Compute action-value directly using approximate critic.

        Calculates :math:`Q^{\\pi}(s, a)` as a target for each model in the
        ensemble. Each value is the minimum among critic predictions.

        Args:
            obs: The observation tensor of shape `(N, *, O)`
            action: The action Tensor of shape `(N, *, A)`

        Returns:
            The action-value tensor of shape `(N, *)`
        """
        unclipped_qval = self._modules["critics"](obs, action)
        return unclipped_qval.min(dim=-1)[0]

    def one_step_action_value_surrogate(self, obs: Tensor, action: Tensor) -> Tensor:
        """Surrogate loss for gradient estimation of action values.

        Args:
            obs: The observation tensor of shape `(N, *, O)`
            action: The action Tensor of shape `(N, *, A)`

        Returns:
            A tensor of shape `(N, *)` for estimating the gradient of the 1-step
            action-value function.
        """
        next_obs, next_obs_logp = self.transition(obs, action)
        next_act, next_act_logp = self._modules["policy"].rsample(next_obs)

        unclipped_qval = self._modules["critics"](next_obs, next_act)
        next_qval, _ = unclipped_qval.min(dim=-1)

        reward = self._env.reward(obs, action, next_obs)
        done = self._env.termination(obs, action, next_obs)

        next_vval = (
            torch.where(done, reward, reward + self.gamma * next_qval)
            - self.alpha * next_act_logp
        )

        if self.grad_estimator == "SF":
            surrogate = next_obs_logp * next_vval.detach()
        elif self.grad_estimator == "PD":
            surrogate = next_vval
        return surrogate

    def transition(self, obs: Tensor, action: Tensor) -> Tuple[Tensor, Tensor]:
        """Compute virtual transition and its log density.

        Args:
            obs: The current state tensor of shape `(N, *, O)`
            action: The action tensor of shape `(N, *, A)` sampled from the
                stochastic policy

        Returns:
            A tuple with the next state tensor of shape `(N, *, O)` and its
            log-likelihood tensor of shape `(N, *)` generated from models in the
            ensemble
        """
        models = self._modules["models"]

        if self.grad_estimator == "SF":
            next_obs, logp = models.sample(obs, action)
        elif self.grad_estimator == "PD":
            next_obs, logp = models.rsample(obs, action)

        return next_obs, logp

    @staticmethod
    def action_gradient_loss(
        action: Tensor, value_target: Tensor, value_pred: Tensor
    ) -> Tensor:
        """Decision-aware model loss based on action gradients.

        Args:
            action: The action tensor of shape `(N, *, A)`
            value_target: The estimated action-value gradient target tensor of
                shape `(N, *)`
            value_pred: The surrogate loss tensor of shape `(N, *)` for action
                gradient estimation of the 1-step action-value

        Returns:
            The loss tensor of shape `(N,)`
        """
        temporal_diff = value_target - value_pred
        (action_gradient,) = grad(temporal_diff.sum(), action, create_graph=True)

        # First compute action gradient norms by reducing along action dimension
        grad_norms = action_gradient.abs().sum(dim=-1)
        # Return mean action gradient loss along batch dimension
        return grad_norms.mean(dim=-1)

    def maximum_likelihood_loss(self, batch: TensorDict) -> Tensor:
        """Model regularization through Maximum Likelihood.

        Args:
            batch: The tensor batch with SampleBatch.CUR_OBS,
                SampleBatch.ACTIONS, and SampleBatch.NEXT_OBS keys

        Returns:
            The loss tensor of shape `(N,)`
        """
        loss, _ = self._loss_mle(batch)
        return loss



"""SoftSVG policy class using PyTorch."""
import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from ray.rllib.utils import override

from raylab.agents.svg import SVGTorchPolicy
from raylab.policy import EnvFnMixin
from raylab.policy.losses import ISSoftVIteration
from raylab.policy.losses import MaximumEntropyDual
from raylab.policy.losses import OneStepSoftSVG
from raylab.pytorch.optim import build_optimizer


class SoftSVGTorchPolicy(SVGTorchPolicy):
    """Stochastic Value Gradients policy for off-policy learning."""

    # pylint:disable=abstract-method

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        assert "target_critic" in self.module, "SoftSVG needs a target Value function!"

        self.loss_actor = OneStepSoftSVG(
            self.module.model.reproduce,
            self.module.actor.reproduce,
            self.module.critic,
        )
        self.loss_actor.gamma = self.config["gamma"]

        self.loss_critic = ISSoftVIteration(
            self.module.critic, self.module.target_critic, self.module.actor.sample,
        )
        self.loss_critic.gamma = self.config["gamma"]

        target_entropy = (
            -action_space.shape[0]
            if self.config["target_entropy"] == "auto"
            else self.config["target_entropy"]
        )
        self.loss_alpha = MaximumEntropyDual(
            self.module.alpha, self.module.actor.sample, target_entropy
        )

    @property
    @override(SVGTorchPolicy)
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.svg.soft import SoftSVGTrainer

        return SoftSVGTrainer.options

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(SVGTorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["torch_optimizer"]
        components = {
            "model": self.module.model,
            "actor": self.module.actor,
            "critic": self.module.critic,
            "alpha": self.module.alpha,
        }

        mapping = {
            name: build_optimizer(module, config[name])
            for name, module in components.items()
        }

        optimizers.update(mapping)
        return optimizers

    @torch.no_grad()
    @override(SVGTorchPolicy)
    def add_truncated_importance_sampling_ratios(self, batch_tensors):
        """Compute and add truncated importance sampling ratios to tensor batch."""
        curr_logp = self.module.actor.log_prob(
            batch_tensors[SampleBatch.CUR_OBS], batch_tensors[SampleBatch.ACTIONS]
        )

        is_ratios = torch.exp(curr_logp - batch_tensors[SampleBatch.ACTION_LOGP])
        _is_ratios = torch.clamp(is_ratios, max=self.config["max_is_ratio"])

        batch_tensors[self.loss_actor.IS_RATIOS] = _is_ratios
        batch_tensors[self.loss_critic.IS_RATIOS] = _is_ratios

        info = {
            "is_ratio_max": is_ratios.max().item(),
            "is_ratio_mean": is_ratios.mean().item(),
            "is_ratio_min": is_ratios.min().item(),
            "cross_entropy": -curr_logp.mean().item(),
        }
        return batch_tensors, info

    @override(SVGTorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        batch_tensors, info = self.add_truncated_importance_sampling_ratios(
            batch_tensors
        )

        alpha = self.module.alpha().item()
        self.loss_critic.alpha = alpha
        self.loss_actor.alpha = alpha

        info.update(self._update_model(batch_tensors))
        info.update(self._update_critic(batch_tensors))
        info.update(self._update_actor(batch_tensors))
        if self.config["target_entropy"] is not None:
            info.update(self._update_alpha(batch_tensors))

        self._update_polyak()
        return info

    def _update_model(self, batch_tensors):
        with self.optimizers.optimize("model"):
            model_loss, info = self.loss_model(batch_tensors)
            model_loss.backward()

        info.update(self.extra_grad_info("model"))
        return info

    def _update_critic(self, batch_tensors):
        with self.optimizers.optimize("critic"):
            value_loss, info = self.loss_critic(batch_tensors)
            value_loss.backward()

        info.update(self.extra_grad_info("critic"))
        return info

    def _update_actor(self, batch_tensors):
        with self.optimizers.optimize("actor"):
            svg_loss, info = self.loss_actor(batch_tensors)
            svg_loss.backward()

        info.update(self.extra_grad_info("actor"))
        return info

    def _update_alpha(self, batch_tensors):
        with self.optimizers.optimize("alpha"):
            alpha_loss, info = self.loss_alpha(batch_tensors)
            alpha_loss.backward()

        info.update(self.extra_grad_info("alpha"))
        return info

    @torch.no_grad()
    def extra_grad_info(self, component):
        """Return gradient statistics for component."""
        fetches = {
            f"grad_norm({component})": nn.utils.clip_grad_norm_(
                self.module[component].parameters(), float("inf")
            ).item()
        }
        return fetches



"""Losses for Model-Aware Policy Optimization."""
from typing import Tuple

import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.policy.modules.actor.policy.stochastic import StochasticPolicy
from raylab.policy.modules.critic.q_value import QValueEnsemble
from raylab.policy.modules.model.stochastic.ensemble import StochasticModelEnsemble
from raylab.utils.annotations import DynamicsFn
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin
from .mixins import UniformModelPriorMixin


class MAPO(EnvFunctionsMixin, UniformModelPriorMixin, Loss):
    """Model-Aware Policy Optimization.

    Args:
        models: ensemble of stochastic models
        actor: stochastic policy
        critics: Q-value estimators

    Attributes:
        gamma: Discount factor
        alpha: Entropy regularization coefficient
    """

    batch_keys = (SampleBatch.CUR_OBS,)
    gamma: float = 0.99
    alpha: float = 0.05

    def __init__(
        self,
        models: StochasticModelEnsemble,
        actor: StochasticPolicy,
        critics: QValueEnsemble,
    ):
        super().__init__()
        modules = nn.ModuleDict()
        modules["models"] = models
        modules["policy"] = actor
        modules["critics"] = critics
        self._modules = modules

    @property
    def initialized(self) -> bool:
        """Whether or not the loss function has all the necessary components."""
        return self._env.initialized

    @property
    def _models(self):
        return self._modules["models"]

    def compile(self):
        self._modules.update({k: torch.jit.script(m) for k, m in self._modules.items()})

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        assert self.initialized, (
            "Environment functions missing. "
            "Did you set reward and termination functions?"
        )
        obs = batch[SampleBatch.CUR_OBS]
        action, action_logp, policy_info = self._generate_action(obs)
        next_obs, obs_logp, dist_params = self.transition(obs, action)
        action_value = self.one_step_action_value_surrogate(
            obs, action, next_obs, obs_logp
        )

        entropy = -action_logp.mean()

        # Important to minimize the negative entropy
        loss = -action_value.mean() - self.alpha * entropy

        stats = {
            "loss(actor)": loss.item(),
            "entropy": entropy.item(),
        }
        stats.update(self.model_dist_info(dist_params))
        stats.update(policy_info)
        return loss, stats

    def _generate_action(self, obs: Tensor) -> Tuple[Tensor, Tensor, StatDict]:
        policy = self._modules["policy"]
        dist_params = policy(obs)
        sample, logp = policy.dist.rsample(dist_params)
        info = {}
        info.update(
            {"policy_mean_" + k: v.mean().item() for k, v in dist_params.items()}
        )
        info.update({"policy_max_" + k: v.max().item() for k, v in dist_params.items()})
        info.update({"policy_min_" + k: v.min().item() for k, v in dist_params.items()})
        return sample, logp, info

    def one_step_action_value_surrogate(
        self, obs: Tensor, action: Tensor, next_obs: Tensor, log_prob: Tensor
    ) -> Tensor:
        """Surrogate loss for gradient estimation of action values.

        Computes 1-step approximation of Q^{\\pi}(s, a) for maximum entropy
        framework.

        Args:
            obs: The current observation
            action: Action taken by the agent
            next_obs: The observation resulting from the applied action
            log_prob: The log-probability of the next observation

        Returns:
            A tensor for estimating the gradient of the 1-step action-value
            function.
        """
        next_act, logp = self._modules["policy"].rsample(next_obs)

        unclipped_qval = self._modules["critics"](next_obs, next_act)
        next_qval, _ = unclipped_qval.min(dim=-1)

        reward = self._env.reward(obs, action, next_obs)
        done = self._env.termination(obs, action, next_obs)

        next_vval = (
            torch.where(done, reward, reward + self.gamma * next_qval)
            - self.alpha * logp
        )

        if self.grad_estimator == "SF":
            surrogate = torch.mean(log_prob * next_vval.detach(), dim=0)
        elif self.grad_estimator == "PD":
            surrogate = torch.mean(next_vval, dim=0)
        return surrogate


class DAPO(EnvFunctionsMixin, Loss):
    """Dynamics-Aware Policy Optimization.

    Computes the 1-step maximum entropy policy loss using a given dynamics
    function.

    Args:
        dynamics_fn: Dynamics function, usually provided by the environment
        actor: Stochastic parameterized policy
        critics: Q-value estimators

    Attributes:
        gamma: Discount factor
        alpha: Entropy regularization coefficient
        grad_estimator: Gradient estimator for expecations ('PD' or 'SF')
    """

    gamma: float = 0.99
    alpha: float = 0.05
    grad_estimator: str = "SF"

    def __init__(
        self, dynamics_fn: DynamicsFn, actor: StochasticPolicy, critics: QValueEnsemble
    ):
        super().__init__()
        self.dynamics_fn = dynamics_fn
        modules = nn.ModuleDict()
        modules["policy"] = actor
        modules["critics"] = critics
        self._modules = modules

    @property
    def initialized(self) -> bool:
        """Whether or not the loss function has all the necessary components."""
        return self._env.initialized

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        assert self.initialized, (
            "Environment functions missing. "
            "Did you set reward and termination functions?"
        )
        obs = batch[SampleBatch.CUR_OBS]
        action, action_logp = self._modules["policy"].rsample(obs)
        next_obs, obs_logp = self.transition(obs, action)
        action_value = self.one_step_action_value_surrogate(
            obs, action, next_obs, obs_logp
        )

        entropy = -action_logp.mean()
        # Important to minimize the negative entropy
        loss = -action_value.mean() - self.alpha * entropy

        stats = {
            "loss(actor)": loss.item(),
            "entropy": entropy.item(),
        }
        return loss, stats

    def transition(self, obs: Tensor, action: Tensor) -> Tuple[Tensor, Tensor]:
        """Compute virtual transition and its log density.

        Args:
            obs: The current state
            action: The action sampled from the stochastic policy

        Returns:
            A tuple with the next state and its log-likelihood generated from
            the dynamics function
        """
        next_obs, logp = self.dynamics_fn(obs, action)
        if self.grad_estimator == "SF":
            next_obs = next_obs.detach()
        return next_obs, logp

    def one_step_action_value_surrogate(
        self, obs: Tensor, action: Tensor, next_obs: Tensor, log_prob: Tensor
    ) -> Tensor:
        """Surrogate loss for gradient estimation of action values.

        Computes 1-step approximation of Q^{\\pi}(s, a) for maximum entropy
        framework.

        Args:
            obs: The current observation
            action: Action taken by the agent
            next_obs: The observation resulting from the applied action
            log_prob: The log-probability of the next observation

        Returns:
            A tensor for estimating the gradient of the 1-step action-value
            function.
        """
        next_act, logp = self._modules["policy"].rsample(next_obs)

        unclipped_qval = self._modules["critics"](next_obs, next_act)
        next_qval, _ = unclipped_qval.min(dim=-1)

        reward = self._env.reward(obs, action, next_obs)
        done = self._env.termination(obs, action, next_obs)

        next_vval = (
            torch.where(done, reward, reward + self.gamma * next_qval)
            - self.alpha * logp
        )

        if self.grad_estimator == "SF":
            surrogate = torch.mean(log_prob * next_vval.detach(), dim=0)
        elif self.grad_estimator == "PD":
            surrogate = torch.mean(next_vval, dim=0)
        return surrogate



# pylint:disable=missing-module-docstring
from raylab.agents import trainer
from raylab.agents.model_based import set_policy_with_env_fn
from raylab.agents.off_policy import OffPolicyTrainer
from raylab.agents.sac.trainer import sac_config

from .policy import DAPOTorchPolicy


@trainer.configure
@trainer.option("losses/", help="Configurations for actor loss function")
@trainer.option(
    "losses/grad_estimator",
    default="PD",
    help="""Gradient estimator for optimizing expectations.

    Possible types include
    SF: score function
    PD: pathwise derivative
    """,
)
@trainer.option(
    "losses/model_samples",
    default=1,
    help="Number of next states to sample from the model when calculating the"
    " model-aware deterministic policy gradient",
)
@trainer.option("module/type", default="SAC")
@sac_config
class DAPOTrainer(OffPolicyTrainer):
    """Single-agent trainer for Dynamics-Aware Policy Optimization"""

    _name = "DAPO"
    _policy = DAPOTorchPolicy

    def _init(self, config, env_creator):
        super()._init(config, env_creator)
        set_policy_with_env_fn(self.workers, fn_type="reward")
        set_policy_with_env_fn(self.workers, fn_type="termination")
        set_policy_with_env_fn(self.workers, fn_type="dynamics")



"""Support for modules with state value functions as critics."""
import torch.nn as nn

from raylab.pytorch.nn import FullyConnected
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge


BASE_CONFIG = {
    "target_vf": True,
    "initializer_options": {"name": "xavier_uniform"},
    "encoder": {"units": (400, 200), "activation": "Tanh"},
}


class StateValueMixin:
    """Adds constructor for modules with state value functions.

    Its is not common to have target value functions in this setting, but the option
    is provided regardless.
    """

    # pylint:disable=too-few-public-methods

    @staticmethod
    def _make_critic(obs_space, action_space, config):
        # pylint:disable=unused-argument
        modules = {}
        config = deep_merge(
            BASE_CONFIG,
            config.get("critic", {}),
            False,
            ["encoder", "initializer_options"],
        )

        def make_vf():
            logits_mod = FullyConnected(
                in_features=obs_space.shape[0], **config["encoder"]
            )
            logits_mod.apply(
                initialize_(
                    activation=config["encoder"]["activation"],
                    **config["initializer_options"]
                )
            )
            value_mod = nn.Linear(logits_mod.out_features, 1)
            return nn.Sequential(logits_mod, value_mod)

        modules["critic"] = make_vf()
        if config["target_vf"]:
            modules["target_critic"] = make_vf()
            modules["target_critic"].load_state_dict(modules["critic"].state_dict())
        return modules



"""Mathematical functions."""
import torch


def cbrt(value):
    """Cube root. Equivalent to torch.pow(value, 1/3), but numerically stable.

    Source: https://github.com/bayesiains/nsf/blob/master/utils/torchutils.py
    """
    return torch.sign(value) * torch.exp(torch.log(torch.abs(value)) / 3.0)



"""Collection of Reinforcement Learning components implementations."""

from .action_value_mixin import ActionValueFunction
from .action_value_mixin import ActionValueMixin
from .deterministic_actor_mixin import DeterministicActorMixin
from .deterministic_actor_mixin import DeterministicPolicy
from .normalizing_flow_actor_mixin import NormalizingFlowActorMixin
from .normalizing_flow_model_mixin import NormalizingFlowModelMixin
from .state_value_mixin import StateValueMixin
from .stochastic_actor_mixin import MaximumEntropyMixin
from .stochastic_actor_mixin import StochasticActorMixin
from .stochastic_actor_mixin import StochasticPolicy
from .stochastic_model_mixin import StochasticModel
from .stochastic_model_mixin import StochasticModelMixin
from .svg_model_mixin import SVGModelMixin

__all__ = [
    "ActionValueMixin",
    "ActionValueFunction",
    "DeterministicActorMixin",
    "DeterministicPolicy",
    "NormalizingFlowActorMixin",
    "NormalizingFlowModelMixin",
    "StateValueMixin",
    "StochasticActorMixin",
    "MaximumEntropyMixin",
    "StochasticPolicy",
    "StochasticModelMixin",
    "StochasticModel",
    "SVGModelMixin",
]



"""
MIT License

Copyright (c) 2019 Conor Durkan, Artur Bekasov, Iain Murray, George Papamakarios

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


Slightly modified from:
https://github.com/bayesiains/nsf/blob/master/utils/torchutils.py
"""
import torch


def create_alternating_binary_mask(features, even=True):
    """
    Creates a binary mask of a given dimension which alternates its masking.
    :param features: Dimension of mask.
    :param even: If True, even values are assigned 1s, odd 0s. If False, vice versa.
    :return: Alternating binary mask of type torch.Tensor.
    """
    mask = torch.zeros(features).byte()
    start = 0 if even else 1
    mask[start::2] += 1
    return mask


def create_mid_split_binary_mask(features):
    """
    Creates a binary mask of a given dimension which splits its masking at the midpoint.
    :param features: Dimension of mask.
    :return: Binary mask split at midpoint of type torch.Tensor
    """
    mask = torch.zeros(features).byte()
    midpoint = features // 2 if features % 2 == 0 else features // 2 + 1
    mask[:midpoint] += 1
    return mask


def create_random_binary_mask(features):
    """
    Creates a random binary mask of a given dimension with half of its entries
    randomly set to 1s.
    :param features: Dimension of mask.
    :return: Binary mask with half of its entries set to 1s, of type torch.Tensor.
    """
    mask = torch.zeros(features).byte()
    weights = torch.ones(features).float()
    num_samples = features // 2 if features % 2 == 0 else features // 2 + 1
    indices = torch.multinomial(
        input=weights, num_samples=num_samples, replacement=False
    )
    mask[indices] += 1
    return mask



"""Abstract base classes for losses."""
from abc import ABCMeta
from abc import abstractmethod
from typing import Dict
from typing import Tuple

from torch import Tensor

from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

__all__ = [
    "Dict",
    "Tuple",
    "Tensor",
    "Loss",
]


class Loss(metaclass=ABCMeta):
    """Base interface for all loss classes.

    Attributes:
        batch_keys: the fiels in the tensor batch which will be accessed when
            called. Needed for converting the appropriate inputs to tensors
            externally.
    """

    batch_keys: Tuple[str, ...]

    @abstractmethod
    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Computes the loss function and stats dict for the given batch.

        Subclasses should override this to implement their respective loss
        functions.

        Args:
            batch: a dictionary of input tensors

        Returns:
            A tuple with the loss tensor (not necessarily scalar) and statistics
        """

    def compile(self):
        """Optimize this loss function's implementation if supported.

        Subclasses should use TorchScript to build a static computation graph,
        although this may not be possible in some cases.
        """



"""Trainer and configuration for MBPO."""
from raylab.agents import trainer
from raylab.agents.model_based import DynaLikeTrainer
from raylab.agents.sac.trainer import sac_config
from raylab.policy.model_based.sampling_mixin import SamplingSpec
from raylab.policy.model_based.training_mixin import TrainingSpec

from .policy import MBPOTorchPolicy

DEFAULT_MODULE = {
    "type": "ModelBasedSAC",
    "model": {
        "network": {"units": (128, 128), "activation": "Swish"},
        "ensemble_size": 7,
        "input_dependent_scale": True,
        "parallelize": True,
        "residual": True,
    },
    "actor": {
        "encoder": {"units": (128, 128), "activation": "Swish"},
        "input_dependent_scale": True,
    },
    "critic": {
        "double_q": True,
        "encoder": {"units": (128, 128), "activation": "Swish"},
    },
    "entropy": {"initial_alpha": 0.05},
}


@trainer.configure
@trainer.option("module", default=DEFAULT_MODULE, override=True)
@trainer.option(
    "torch_optimizer/models",
    default={"type": "Adam", "lr": 3e-4, "weight_decay": 0.0001},
)
@trainer.option(
    "model_training", default=TrainingSpec().to_dict(), help=TrainingSpec.__doc__
)
@trainer.option(
    "model_warmup",
    default=TrainingSpec().to_dict(),
    help="""Specifications for model warm-up.

    Same configurations as 'model_training'.
    """,
)
@trainer.option(
    "model_sampling", default=SamplingSpec().to_dict(), help=SamplingSpec.__doc__
)
@trainer.option("model_rollouts", 20, override=True)
@trainer.option("learning_starts", 5000, override=True)
@trainer.option("train_batch_size", 512, override=True)
@trainer.option("compile_policy", True, override=True)
@sac_config
class MBPOTrainer(DynaLikeTrainer):
    """Model-based trainer using SAC for policy improvement."""

    _name = "MBPO"
    _policy = MBPOTorchPolicy



"""Customized Linear modules."""
import torch
import torch.nn as nn
import torch.nn.functional as F
from ray.rllib.utils import override

from raylab.pytorch.nn.init import initialize_


class MaskedLinear(nn.Linear):
    """Linear module with a configurable mask on the weights."""

    def __init__(self, in_features, out_features, bias=True):
        super().__init__(in_features, out_features, bias)
        self.register_buffer("mask", torch.ones(out_features, in_features))

    @torch.jit.export
    def set_mask(self, mask):
        """Update mask tensor."""
        self.mask.data.copy_(mask)

    @override(nn.Linear)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return F.linear(inputs, self.mask * self.weight, self.bias)


class NormalizedLinear(nn.Module):
    """Enforces a norm constraint on outputs."""

    __constants__ = {"beta"}

    def __init__(self, *args, beta, **kwargs):
        super().__init__()
        self.linear = nn.Linear(*args, **kwargs)
        self.beta = beta
        self.apply(initialize_("xavier_uniform", activation="tanh"))

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        vec = self.linear(inputs)
        norms = vec.norm(p=1, dim=-1, keepdim=True)
        normalized = vec * self.linear.out_features * self.beta / norms
        return torch.where(
            norms / self.linear.out_features > self.beta, normalized, vec
        )



# pylint:disable=missing-docstring
import torch
import torch.nn as nn
from ray.rllib.utils import override


class GaussianNoise(nn.Module):
    """Neural network module adding gaussian i.i.d. noise to inputs."""

    __constants__ = {"scale"}

    def __init__(self, scale):
        super().__init__()
        self.scale = scale

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return inputs + torch.randn_like(inputs) * self.scale



"""Actor-Critic architecture used in most on-policy PG algorithms."""
from ray.rllib.utils import merge_dicts

from .abstract import AbstractActorCritic
from .mixins import StateValueMixin
from .mixins import StochasticActorMixin


BASE_CONFIG = {
    "actor": {
        "encoder": {"units": (32, 32), "activation": "Tanh"},
        "input_dependent_scale": False,
    },
    "critic": {
        "encoder": {"units": (32, 32), "activation": "Tanh"},
        "target_vf": False,
    },
}


class OnPolicyActorCritic(
    StochasticActorMixin, StateValueMixin, AbstractActorCritic,
):
    """Actor-Critic module with stochastic actor and state-value critics."""

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        super().__init__(obs_space, action_space, merge_dicts(BASE_CONFIG, config))



"""Losses for model aware action gradient estimation."""
from dataclasses import dataclass
from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.policy.modules.actor.policy.deterministic import DeterministicPolicy
from raylab.policy.modules.critic.q_value import QValueEnsemble
from raylab.policy.modules.model.stochastic.ensemble import StochasticModelEnsemble
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin
from .mixins import UniformModelPriorMixin


@dataclass
class MAGEModules:
    """Necessary modules for MAGE loss.

    Attributes:
        critics: Q-value estimators
        target_critics: Q-value estimators for the next state
        policy: deterministic policy for current state
        target_policy: deterministic policy for next state
        models: ensemble of stochastic models
    """

    critics: QValueEnsemble
    target_critics: QValueEnsemble
    policy: DeterministicPolicy
    target_policy: DeterministicPolicy
    models: StochasticModelEnsemble


class MAGE(EnvFunctionsMixin, UniformModelPriorMixin, Loss):
    """Loss function for Model-based Action-Gradient-Estimator.

    Args:
        modules: necessary modules for MAGE loss

    Attributes:
        gamma: discount factor
        lambda: weighting factor for TD-error regularization
    """

    batch_keys = (SampleBatch.CUR_OBS,)
    gamma: float = 0.99
    lambda_: float = 0.05

    def __init__(self, modules: MAGEModules):
        super().__init__()
        self._modules = nn.ModuleDict(
            dict(
                critics=modules.critics,
                target_critics=modules.target_critics,
                policy=modules.policy,
                target_policy=modules.target_policy,
                models=modules.models,
            )
        )
        self._rng = np.random.default_rng()

    @property
    def initialized(self) -> bool:
        """Whether or not the loss function has all the necessary components."""
        return self._env.initialized

    @property
    def grad_estimator(self):
        """Gradient estimator for expecations."""
        return "PD"

    @property
    def model_samples(self):
        """Number of next states to draw from the model"""
        return 1

    @property
    def _models(self) -> StochasticModelEnsemble:
        return self._modules["models"]

    def compile(self):
        self._modules.update({k: torch.jit.script(v) for k, v in self._modules.items()})

    def transition(self, obs, action):
        next_obs, _, dist_params = super().transition(obs, action)
        # Squeeze the model sample_shape dimension
        return next_obs.squeeze(dim=0), dist_params

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        assert self.initialized, (
            "Environment functions missing. "
            "Did you set reward, termination, and dynamics functions?"
        )

        obs = batch[SampleBatch.CUR_OBS]
        action = self._modules["policy"](obs)
        next_obs, dist_params = self.transition(obs, action)

        delta = self.temporal_diff_error(obs, action, next_obs)
        grad_loss = self.gradient_loss(delta, action)
        td_reg = self.temporal_diff_loss(delta)
        loss = grad_loss + self.lambda_ * td_reg

        info = {
            "loss(critics)": loss.item(),
            "loss(MAGE)": grad_loss.item(),
            "loss(TD)": td_reg.item(),
        }
        info.update(self.model_dist_info(dist_params))
        return loss, info

    def temporal_diff_error(
        self, obs: Tensor, action: Tensor, next_obs: Tensor,
    ) -> Tensor:
        """Returns the temporal difference error with clipped action values."""
        values = torch.cat([m(obs, action) for m in self._modules["critics"]], dim=-1)

        reward = self._env.reward(obs, action, next_obs)
        done = self._env.termination(obs, action, next_obs)
        next_action = self._modules["target_policy"](next_obs)
        next_values, _ = torch.cat(
            [m(next_obs, next_action) for m in self._modules["target_critics"]], dim=-1
        ).min(dim=-1)
        targets = torch.where(done, reward, reward + self.gamma * next_values)
        targets = targets.unsqueeze(-1).expand_as(values)

        return targets - values

    @staticmethod
    def gradient_loss(delta: Tensor, action: Tensor) -> Tensor:
        """Returns the action gradient loss for the Q-value function."""
        (action_gradient,) = torch.autograd.grad(delta.sum(), action, create_graph=True)
        return action_gradient.abs().sum(dim=-1).mean()

    @staticmethod
    def temporal_diff_loss(delta: Tensor) -> Tensor:
        """Returns the temporal difference loss for the Q-value function."""
        return delta.abs().sum(dim=-1).mean()



"""Trainer and configuration for TRPO."""
from ray.rllib.optimizers import SyncSamplesOptimizer
from ray.rllib.utils import override

from raylab.agents import Trainer
from raylab.agents import trainer

from .policy import TRPOTorchPolicy

LINESEARCH_OPTIONS = {
    "accept_ratio": 0.1,
    "backtrack_ratio": 0.8,
    "max_backtracks": 15,
    "atol": 1e-7,
}


@trainer.configure
@trainer.option("delta", 0.01, help="Trust region constraint")
@trainer.option(
    "fvp_samples",
    10,
    help="Number of actions to sample per state for Fisher vector-product calculation",
)
@trainer.option("lambda", 0.97, help=r"For GAE(\gamma, \lambda)")
@trainer.option("val_iters", 80, help="Number of iterations to fit value function")
@trainer.option(
    "torch_optimizer",
    {"type": "Adam", "lr": 1e-3},
    override=True,
    help="Options for the critic's optimizer",
)
@trainer.option("use_gae", True, help="Whether to use Generalized Advantage Estimation")
@trainer.option("cg_iters", 10, help="Number of iterations for Conjugate Gradient")
@trainer.option(
    "cg_damping",
    1e-3,
    help="Damping factor to avoid singular matrix multiplication in Conjugate Gradient",
)
@trainer.option(
    "line_search",
    True,
    help="""
    Whether to use a line search to calculate policy update.
    Effectively turns TRPO into Natural PG when turned off.
    """,
)
@trainer.option("line_search_options", LINESEARCH_OPTIONS)
@trainer.option("module/type", "OnPolicyActorCritic-v0")
@trainer.option(
    "exploration_config/type",
    "raylab.utils.exploration.StochasticActor",
    override=True,
)
class TRPOTrainer(Trainer):
    """Single agent trainer for TRPO."""

    _name = "TRPO"
    _policy = TRPOTorchPolicy

    # pylint:disable=attribute-defined-outside-init

    @override(Trainer)
    def _init(self, config, env_creator):
        self._validate_config(config)
        self.workers = self._make_workers(
            env_creator, self._policy, config, num_workers=config["num_workers"]
        )
        self.optimizer = SyncSamplesOptimizer(
            self.workers, train_batch_size=config["train_batch_size"]
        )

    @override(Trainer)
    def _train(self):
        init_timesteps = self.optimizer.num_steps_sampled
        timesteps_per_iteration = max(self.config["timesteps_per_iteration"], 2)

        while (
            self.optimizer.num_steps_sampled - init_timesteps < timesteps_per_iteration
        ):
            learner_stats = self.optimizer.step()

        res = self.collect_metrics()
        timesteps = self.optimizer.num_steps_sampled - init_timesteps
        res.update(timesteps_this_iter=timesteps, learner_stats=learner_stats)
        return res



"""Custom Replay Buffers based on RLlibs's implementation."""
import random
import sys
from dataclasses import dataclass

import numpy as np
from gym.spaces import Space
from ray.rllib import SampleBatch
from ray.rllib.optimizers.replay_buffer import ReplayBuffer as _ReplayBuffer
from ray.rllib.utils import override
from ray.rllib.utils.compression import unpack_if_needed


@dataclass
class ReplayField:
    """Storage specification for ReplayBuffer data."""

    name: str
    shape: tuple = ()
    dtype: np.dtype = np.float32


class ListReplayBuffer(_ReplayBuffer):
    """Replay buffer as a list of tuples.

    Returns a SampleBatch object when queried for samples.

    Args:
        size: max number of transitions to store in the buffer. When the buffer
            overflows, the old memories are dropped.

    Attributes:
        fields (:obj:`tuple` of :obj:`ReplayField`): storage fields
            specification
    """

    def __init__(self, size: int):
        super().__init__(size)
        self.fields = (
            ReplayField(SampleBatch.CUR_OBS),
            ReplayField(SampleBatch.ACTIONS),
            ReplayField(SampleBatch.REWARDS),
            ReplayField(SampleBatch.NEXT_OBS),
            ReplayField(SampleBatch.DONES),
        )

    def add_fields(self, *fields: ReplayField):
        """Add fields to the replay buffer and build the corresponding storage."""
        new_names = {f.name for f in fields}
        assert len(new_names) == len(fields), "Field names must be unique"

        conflicts = new_names.intersection({f.name for f in self.fields})
        assert not conflicts, f"{conflicts} are already in buffer"

        self.fields = self.fields + fields

    @override(_ReplayBuffer)
    def add(self, row: dict):  # pylint:disable=arguments-differ
        """Add a row from a SampleBatch to storage.

        Args:
            row: sample batch row as returned by SampleBatch.rows
        """
        data = tuple(row[f.name] for f in self.fields)
        self._num_added += 1

        if self._next_idx >= len(self._storage):
            self._storage.append(data)
            self._est_size_bytes += sum(sys.getsizeof(d) for d in data)
        else:
            self._storage[self._next_idx] = data
        if self._next_idx + 1 >= self._maxsize:
            self._eviction_started = True
        self._next_idx = (self._next_idx + 1) % self._maxsize
        if self._eviction_started:
            self._evicted_hit_stats.push(self._hit_count[self._next_idx])
            self._hit_count[self._next_idx] = 0

    @override(_ReplayBuffer)
    def _encode_sample(self, idxes):
        sample = []
        for i in idxes:
            sample.append(self._storage[i])
            self._hit_count[i] += 1

        obses_t, actions, rewards, obses_tp1, dones, *extras = zip(*sample)

        obses_t = [np.array(unpack_if_needed(o), copy=False) for o in obses_t]
        actions = [np.array(a, copy=False) for a in actions]
        obses_tp1 = [np.array(unpack_if_needed(o), copy=False) for o in obses_tp1]

        return tuple(
            map(np.array, [obses_t, actions, rewards, obses_tp1, dones] + extras)
        )

    @override(_ReplayBuffer)
    def sample(self, batch_size: int) -> SampleBatch:
        idxes = random.choices(range(len(self._storage)), k=batch_size)
        return self.sample_with_idxes(idxes)

    @override(_ReplayBuffer)
    def sample_with_idxes(self, idxes: np.ndarray) -> SampleBatch:
        self._num_sampled += len(idxes)
        data = self._encode_sample(idxes)
        return SampleBatch(dict(zip([f.name for f in self.fields], data)))

    def all_samples(self) -> SampleBatch:
        """All transitions stored in buffer."""
        return self.sample_with_idxes(range(len(self)))


class NumpyReplayBuffer:
    """Replay buffer as a dict of ndarrays.

    Returns a SampleBatch object when queried for samples.

    Args:
        obs_space: observation space
        action_space: action space
        size: max number of transitions to store in the buffer.
            When the bufferoverflows the old memories are dropped.

    Attributes:
        fields (:obj:`tuple` of :obj:`ReplayField`): storage fields
            specification
    """

    def __init__(self, obs_space: Space, action_space: Space, size: int):
        # pylint:disable=too-many-arguments
        self._maxsize = size
        self.fields = (
            ReplayField(
                SampleBatch.CUR_OBS, shape=obs_space.shape, dtype=obs_space.dtype
            ),
            ReplayField(
                SampleBatch.ACTIONS, shape=action_space.shape, dtype=action_space.dtype
            ),
            ReplayField(SampleBatch.REWARDS, shape=(), dtype=np.float32),
            ReplayField(
                SampleBatch.NEXT_OBS, shape=obs_space.shape, dtype=obs_space.dtype
            ),
            ReplayField(SampleBatch.DONES, shape=(), dtype=np.bool),
        )
        self._storage = {}
        self._build_buffers(*self.fields)
        self._next_idx = 0
        self._curr_size = 0
        self._rng = np.random.default_rng()

    def __len__(self) -> int:
        return self._curr_size

    def _build_buffers(self, *fields: ReplayField):
        storage = self._storage
        size = self._maxsize
        for field in fields:
            storage[field.name] = np.empty((size,) + field.shape, dtype=field.dtype)

    def seed(self, seed: int = None):
        """Seed the random number generator for sampling minibatches."""
        self._rng = np.random.default_rng(seed)

    def add_fields(self, *fields: ReplayField):
        """Add fields to the replay buffer and build the corresponding storage."""
        new_names = {f.name for f in fields}
        assert len(new_names) == len(fields), "Field names must be unique"

        conflicts = new_names.intersection({f.name for f in self.fields})
        assert not conflicts, f"{conflicts} are already in buffer"

        self.fields = self.fields + fields
        self._build_buffers(*fields)

    def add(self, row: dict):  # pylint:disable=arguments-differ
        """Add a row from a SampleBatch to storage.

        Args:
            row: sample batch row as returned by SampleBatch.rows().
                Must have the same keys as the field names in the buffer.
        """
        for field in self.fields:
            self._storage[field.name][self._next_idx] = row[field.name]

        self._next_idx = (self._next_idx + 1) % self._maxsize
        self._curr_size += 1 if self._curr_size < self._maxsize else 0

    def sample(self, batch_size: int) -> SampleBatch:
        """Transition batch uniformly sampled with replacement."""
        return self.sample_with_idxes(self.sample_idxes(batch_size))

    def sample_idxes(self, batch_size: int) -> np.ndarray:
        """Get random transition indexes uniformly sampled with replacement."""
        return self._rng.integers(self._curr_size, size=batch_size)

    def sample_with_idxes(self, idxes: np.ndarray) -> SampleBatch:
        """Transition batch corresponding with the given indexes."""
        batch = {k: self._storage[k][idxes] for k in (f.name for f in self.fields)}
        return SampleBatch(batch)

    def all_samples(self) -> SampleBatch:
        """All stored transitions."""
        return SampleBatch(
            {k: self._storage[k][: len(self)] for k in (f.name for f in self.fields)}
        )



"""Actor-Critic architecture used in Trust-Region Policy Optimization (TRPO)."""
from typing import Dict

import gym.spaces as spaces
import torch
import torch.nn as nn
from ray.rllib.utils import merge_dicts
from ray.rllib.utils import override

import raylab.policy.modules.networks as networks
from raylab.pytorch.nn import FullyConnected
from raylab.pytorch.nn.distributions import flows
from raylab.pytorch.nn.distributions import Independent
from raylab.pytorch.nn.distributions import Normal
from raylab.pytorch.nn.distributions import TransformedDistribution
from raylab.pytorch.nn.distributions.flows import CompositeTransform
from raylab.pytorch.nn.distributions.flows import TanhSquashTransform
from raylab.pytorch.nn.init import initialize_

from .abstract import AbstractActorCritic
from .mixins import StateValueMixin
from .mixins import StochasticActorMixin
from .mixins import StochasticPolicy


# Defaults for Hopper-v1
BASE_CONFIG = {
    "actor": {"num_flows": 4, "hidden_size": 3},
    "critic": {
        "initializer_options": {"name": "normal", "std": 1.0},
        "encoder": {"units": (32, 32), "activation": "Tanh"},
        "target_vf": False,
    },
}


class TRPOTang2018(StochasticActorMixin, StateValueMixin, AbstractActorCritic):
    """Actor-Critic module with stochastic actor and state-value critics.

    Stochastic policy architecture used in
    http://arxiv.org/abs/1809.10326
    """

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        super().__init__(obs_space, action_space, merge_dicts(BASE_CONFIG, config))

    @override(StochasticActorMixin)
    def _make_actor(self, obs_space, action_space, config):
        config = config["actor"]
        assert isinstance(
            action_space, spaces.Box
        ), f"Normalizing Flows incompatible with action space type {type(action_space)}"

        # PARAMS MODULE ================================================================
        params_module = StateNormalParams(obs_space, action_space)

        # FLOW MODULES =================================================================
        obs_size = obs_space.shape[0]
        act_size = action_space.shape[0]

        def transform_net_fn(in_features, out_features):
            return networks.MLP(
                in_features,
                out_features,
                hidden_features=config["hidden_size"],
                num_blocks=2,
                activation="ReLU",
            )

        masks = [
            flows.masks.create_alternating_binary_mask(act_size, bool(i % 2))
            for i in range(config["num_flows"])
        ]
        couplings = [flows.AffineCouplingTransform(m, transform_net_fn) for m in masks]
        add_state = AddStateFlow(obs_size, act_size)
        squash = TanhSquashTransform(
            low=torch.as_tensor(action_space.low),
            high=torch.as_tensor(action_space.high),
            event_dim=1,
        )
        transforms = couplings[:1] + [add_state] + couplings[1:] + [squash]
        dist_module = TransformedDistribution(
            base_dist=Independent(Normal(), reinterpreted_batch_ndims=1),
            transform=CompositeTransform(transforms),
        )

        return {"actor": StochasticPolicy(params_module, dist_module)}


class StateNormalParams(nn.Module):
    """Maps observations to standard normal parameters and forwards observations."""

    def __init__(self, obs_space, action_space):
        super().__init__()
        self.obs_dim = len(obs_space.shape)
        self.act_shape = action_space.shape

    @override(nn.Module)
    def forward(self, obs):  # pylint:disable=arguments-differ
        shape = obs.shape[: -self.obs_dim] + self.act_shape
        return {"loc": torch.zeros(shape), "scale": torch.ones(shape), "state": obs}


class AddStateFlow(flows.ConditionalTransform):
    """Incorporates state information by adding a state embedding."""

    def __init__(self, obs_size, act_size):
        super().__init__(event_dim=1)
        self.state_encoder = MLP(obs_size, act_size, hidden_size=64)

    @override(flows.ConditionalTransform)
    def encode(self, inputs, params: Dict[str, torch.Tensor]):
        encoded = self.state_encoder(params["state"])
        return inputs + encoded, torch.zeros(inputs.shape[: -self.event_dim])

    @override(flows.ConditionalTransform)
    def decode(self, inputs, params: Dict[str, torch.Tensor]):
        encoded = self.state_encoder(params["state"])
        return inputs - encoded, torch.zeros(inputs.shape[: -self.event_dim])


class MLP(nn.Module):
    # pylint:disable=missing-docstring
    def __init__(self, in_size, out_size, layer_norm=True, hidden_size=3):
        super().__init__()
        fully_connected = FullyConnected(
            in_size, units=(hidden_size,) * 2, activation="ReLU", layer_norm=layer_norm,
        )
        linear = nn.Linear(fully_connected.out_features, out_size)
        linear.apply(initialize_("uniform", a=-3e-3, b=3e-3))
        self.net = nn.Sequential(fully_connected, linear)

    def forward(self, inputs):  # pylint:disable=arguments-differ
        return self.net(inputs)



# pylint:disable=missing-docstring
import torch.nn as nn

from .linear import NormalizedLinear
from .tanh_squash import TanhSquash


class ActionOutput(nn.Sequential):
    """Neural network module mapping inputs to actions in specified range."""

    __constants__ = {"in_features", "out_features"}

    def __init__(self, in_features, action_low, action_high, *, beta):
        self.in_features = in_features
        self.out_features = action_low.shape[-1]
        super().__init__(
            NormalizedLinear(self.in_features, self.out_features, beta=beta),
            TanhSquash(action_low, action_high),
        )



"""Collection of custom RLlib Policy classes."""

from .kl_coeff_mixin import AdaptiveKLCoeffMixin
from .model_based import EnvFnMixin
from .model_based import ModelSamplingMixin
from .model_based import ModelTrainingMixin
from .optimizer_collection import OptimizerCollection
from .torch_policy import TorchPolicy

__all__ = [
    "AdaptiveKLCoeffMixin",
    "EnvFnMixin",
    "ModelSamplingMixin",
    "ModelTrainingMixin",
    "TorchPolicy",
    "OptimizerCollection",
]



"""Model-based Soft Actor-Critic architecture."""

from .abstract import AbstractModelActorCritic
from .mixins import ActionValueMixin
from .mixins import MaximumEntropyMixin
from .mixins import StochasticActorMixin
from .mixins import StochasticModelMixin


# pylint:disable=abstract-method,too-many-ancestors
class ModelBasedSAC(
    StochasticModelMixin,
    StochasticActorMixin,
    ActionValueMixin,
    MaximumEntropyMixin,
    AbstractModelActorCritic,
):
    """
    Module architecture with stochastic actor and model, action-value function, and
    entropy coefficient.
    """



"""Distribution transforms as PyTorch modules compatible with TorchScript."""
from typing import Dict

import torch
import torch.nn as nn
from ray.rllib.utils import override

from .utils import sum_rightmost


class Transform(nn.Module):
    """A diffeomorphism.

    Transforms are differentiable bijections with tractable Jacobians.
    All transforms map samples from a latent space to another (f(z) -> x)
    Use the `reverse` flag to invert the transformation (f^{-1}(x) -> z).
    """

    params: Dict[str, torch.Tensor]

    def __init__(self, *, cond_transform=None, params=None, event_dim=0):
        super().__init__()
        self.event_dim = (
            event_dim if cond_transform is None else cond_transform.event_dim
        )
        self.cond_transform = cond_transform
        self.params = params or {}
        for name, param in self.params.items():
            if isinstance(param, nn.Parameter):
                self.register_parameter(name, param)
            else:
                self.register_buffer(name, param)

    @override(nn.Module)
    def forward(self, inputs, reverse: bool = False):  # pylint:disable=arguments-differ
        return self.decode(inputs) if reverse else self.encode(inputs)

    def encode(self, inputs):
        """
        Computes the transform `z => x` and the log det jacobian `log |dz/dx|`
        """

        return self.cond_transform.encode(inputs, self.params)

    def decode(self, inputs):
        """
        Inverts the transform `x => z` and the log det jacobian `log |dx/dz|`,
        or `- log |dz/dx|`.
        """

        return self.cond_transform.decode(inputs, self.params)


class ConditionalTransform(nn.Module):
    """A Transform conditioned on some external variable(s)."""

    def __init__(self, *, transform=None, event_dim=0):
        super().__init__()
        self.event_dim = event_dim if transform is None else transform.event_dim
        self.transform = transform

    @override(nn.Module)
    def forward(self, inputs, params: Dict[str, torch.Tensor], reverse: bool = False):
        # pylint:disable=arguments-differ
        return self.decode(inputs, params) if reverse else self.encode(inputs, params)

    def encode(self, inputs, params: Dict[str, torch.Tensor]):
        """
        Computes the transform `(z, y) => x`.
        """
        # pylint:disable=unused-argument
        return self.transform.encode(inputs)

    def decode(self, inputs, params: Dict[str, torch.Tensor]):
        """
        Inverts the transform `(x, y) => z`.
        """
        # pylint:disable=unused-argument
        return self.transform.decode(inputs)


class InverseTransform(ConditionalTransform):
    """Invert the transform, effectively swapping the encoding/decoding directions."""

    def __init__(self, transform):
        super().__init__(event_dim=transform.event_dim)
        self.transform = (
            ConditionalTransform(transform=transform)
            if isinstance(transform, Transform)
            else transform
        )

    @override(ConditionalTransform)
    def encode(self, inputs, params: Dict[str, torch.Tensor]):

        return self.transform.decode(inputs, params)

    @override(ConditionalTransform)
    def decode(self, inputs, params: Dict[str, torch.Tensor]):

        return self.transform.encode(inputs, params)


class CompositeTransform(ConditionalTransform):
    # pylint:disable=missing-docstring

    def __init__(self, transforms, event_dim=None):
        event_dim = event_dim or max(t.event_dim for t in transforms)
        super().__init__(event_dim=event_dim)
        assert self.event_dim >= max(t.event_dim for t in transforms), (
            "CompositeTransform cannot have an event_dim smaller than any "
            "of its components'"
        )
        transforms = self.unpack(transforms)
        self.transforms = nn.ModuleList(transforms)
        self.inv_transforms = nn.ModuleList(transforms[::-1])

    @staticmethod
    def unpack(transforms):
        """Recursively unfold CompositeTransforms in a list."""
        result = []
        for trans in transforms:
            if isinstance(trans, CompositeTransform):
                result.extend(trans.unpack(trans.transforms))
            elif isinstance(trans, Transform):
                result += [ConditionalTransform(transform=trans)]
            else:
                result += [trans]
        return result

    @override(ConditionalTransform)
    def encode(self, inputs, params: Dict[str, torch.Tensor]):
        out = inputs
        log_abs_det_jacobian = 0.0
        for transform in self.transforms:
            out, log_det = transform(out, params, reverse=False)
            log_abs_det_jacobian += sum_rightmost(
                log_det, self.event_dim - transform.event_dim
            )
        return out, log_abs_det_jacobian

    @override(ConditionalTransform)
    def decode(self, inputs, params: Dict[str, torch.Tensor]):
        out = inputs
        log_abs_det_jacobian = 0.0
        for transform in self.inv_transforms:
            out, log_det = transform(out, params, reverse=True)
            log_abs_det_jacobian += sum_rightmost(
                log_det, self.event_dim - transform.event_dim
            )
        return out, log_abs_det_jacobian



"""Modularized losses/procedures for RL algorithms."""

from .cdq_learning import ClippedDoubleQLearning
from .cdq_learning import DynaSoftCDQLearning
from .cdq_learning import SoftCDQLearning
from .isfv_iteration import ISFittedVIteration
from .isfv_iteration import ISSoftVIteration
from .mage import MAGE
from .mapo import DAPO
from .mapo import MAPO
from .maximum_entropy import MaximumEntropyDual
from .mle import MaximumLikelihood
from .mle import ModelEnsembleMLE
from .paml import SPAML
from .policy_gradient import ActionDPG
from .policy_gradient import DeterministicPolicyGradient
from .policy_gradient import ReparameterizedSoftPG
from .svg import OneStepSoftSVG
from .svg import OneStepSVG
from .svg import TrajectorySVG


__all__ = [
    "ClippedDoubleQLearning",
    "ActionDPG",
    "DeterministicPolicyGradient",
    "ISFittedVIteration",
    "ISSoftVIteration",
    "MaximumEntropyDual",
    "MAGE",
    "DAPO",
    "MAPO",
    "MaximumLikelihood",
    "ModelEnsembleMLE",
    "PAML",
    "OneStepSVG",
    "OneStepSoftSVG",
    "ReparameterizedSoftPG",
    "SoftCDQLearning",
    "DynaSoftCDQLearning",
    "TrajectorySVG",
]



"""A simple family of diffeomorphisms."""
import torch
from ray.rllib.utils import override
from torch import nn

from .abstract import Transform


class AffineConstantFlow(Transform):
    """
    Scales + Shifts the flow by (learned) constants per dimension.
    In NICE paper there is a Scaling layer which is a special case of
    this where t is None
    """

    def __init__(self, shape, scale=True, shift=True, **kwargs):
        super().__init__(**kwargs)
        if scale:
            self.scale = nn.Parameter(torch.randn(shape))
        else:
            self.register_buffer("scale", torch.zeros(shape))
        if shift:
            self.loc = nn.Parameter(torch.randn(shape))
        else:
            self.register_buffer("loc", torch.zeros(shape))

    @override(Transform)
    def encode(self, inputs):
        out = inputs * torch.exp(self.scale) + self.loc
        # log |dy/dx| = log |torch.exp(scale)| = scale
        log_abs_det_jacobian = self.scale
        return out, log_abs_det_jacobian

    @override(Transform)
    def decode(self, inputs):
        out = (inputs - self.loc) * torch.exp(-self.scale)
        # log |dx/dy| = - log |dy/dx| = - scale
        log_abs_det_jacobian = -self.scale
        return out, log_abs_det_jacobian


class ActNorm(Transform):
    """
    Really an AffineConstantFlow but with a data-dependent initialization,
    where on the very first batch we clever initialize the s,t so that the output
    is unit gaussian. As described in Glow paper.
    """

    def __init__(self, affine_const):
        super().__init__(event_dim=0)
        self.affine_const = affine_const
        self.data_dep_init_done = not (
            isinstance(self.affine_const.scale, nn.Parameter)
            and isinstance(self.affine_const.loc, nn.Parameter)
        )

    @override(Transform)
    def encode(self, inputs):
        # first batch is used for init
        if not self.data_dep_init_done:
            scale = self.affine_const.scale
            loc = self.affine_const.loc

            # pylint:disable=unnecessary-comprehension
            dims = [i for i in range(inputs.dim() - self.event_dim)]
            # pylint:enable=unnecessary-comprehension
            std = -inputs.std(dim=dims).log().detach()
            scale.data.copy_(torch.where(torch.isnan(std), scale, std))
            mean = -torch.mean(inputs * scale.exp(), dim=dims).detach()
            loc.data.copy_(mean)

            self.data_dep_init_done = True
        return self.affine_const.encode(inputs)

    @override(Transform)
    def decode(self, inputs):
        return self.affine_const.decode(inputs)



"""Utilities for manipulating neural network modules."""
import torch
import torch.nn as nn

from .modules.utils import get_activation

__all__ = [
    "get_activation",
    "update_polyak",
    "perturb_params",
]


def update_polyak(from_module: nn.Module, to_module: nn.Module, polyak: float):
    """Update parameters between modules by polyak averaging.

    Args:
        from_module: Module whose parameters are targets.
        to_module: Module whose parameters are updated towards the targets.
        polyak: Averaging factor. The higher it is, the slower the parameters
            are updated.
    """
    for source, target in zip(from_module.parameters(), to_module.parameters()):
        target.data.mul_(polyak).add_(source.data, alpha=1 - polyak)


def perturb_params(target: nn.Module, origin: nn.Module, stddev: float):
    """Set the parameters of a module to a noisy version of another's.

    Loads state dict from the origin module and perturbs the parameters of the
    target module. Layer normalization parameters are ignored.

    Args:
        target: the module to perturb
        origin: the module to copy from
        stddev: the gaussian standard deviation of the noise added to the origin
            model's parameters
    """
    target.load_state_dict(origin.state_dict())

    layer_norms = (m for m in target.modules() if isinstance(m, nn.LayerNorm))
    layer_norm_params = set(p for m in layer_norms for p in m.parameters())
    to_perturb = (p for p in target.parameters() if p not in layer_norm_params)

    for param in to_perturb:
        param.data.add_(torch.randn_like(param) * stddev)



"""
Reference:

Masked Autoregressive Flow for Density Estimation, Papamakarios et al. May 2017
https://arxiv.org/abs/1705.07057

Improved Variational Inference with Inverse Autoregressive Flow, Kingma et al June 2016
https://arxiv.org/abs/1606.04934
(IAF)
"""
import torch
import torch.nn as nn

from raylab.pytorch.nn import MADE

from .abstract import Transform


class ARMLP(nn.Module):
    """A 4-layer auto-regressive MLP, wrapper around MADE net."""

    def __init__(self, in_features, out_features, hidden_dim):
        super().__init__()
        self.net = MADE(
            in_features, (hidden_dim,) * 3, out_features, natural_ordering=True,
        )

    def forward(self, inputs):  # pylint:disable=arguments-differ
        return self.net(inputs)


class MAF(Transform):
    """Masked Autoregressive Flow that uses a MADE-style network for fast forward."""

    def __init__(self, size, parity, net_class=ARMLP, hidden_size=24, **kwargs):
        # prevent further reduction in superclass
        super().__init__(event_dim=0, **kwargs)
        self.size = size
        self.net = net_class(size, size * 2, hidden_size)
        self.parity = parity

    def encode(self, inputs):
        # we have to decode the x one at a time, sequentially
        out = torch.empty_like(inputs)
        log_abs_det_jacobian = torch.zeros(inputs.shape[:-1])
        inputs = inputs.flip(-1) if self.parity else inputs
        for idx in range(self.size):
            # clone to avoid in-place op errors if using IAF
            scale_shift = self.net(out.clone())
            scale, shift = scale_shift.split(self.size, dim=-1)
            out[..., idx] = (inputs[..., idx] - shift[..., idx]) * torch.exp(
                -scale[..., idx]
            )

            log_abs_det_jacobian += -scale[..., idx]
        return out, log_abs_det_jacobian

    def decode(self, inputs):
        # here we see that we are evaluating all of out in parallel,
        # so density estimation will be fast
        scale_shift = self.net(inputs)
        scale, shift = scale_shift.split(self.size, dim=-1)
        out = inputs * torch.exp(scale) + shift
        # reverse order, so if we stack MAFs correct things happen
        out = out.flip(-1) if self.parity else out

        log_abs_det_jacobian = torch.sum(scale, dim=-1)
        return out, log_abs_det_jacobian


class IAF(MAF):
    """
    Reverse the flow, giving an Inverse Autoregressive Flow (IAF) instead,
    where sampling will be fast but density estimation slow.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encode, self.decode = self.decode, self.encode



"""Soft Actor-Critic with Dyna-like data augmentation for critic learning."""
from raylab.agents import trainer
from raylab.agents.model_based import ModelBasedTrainer
from raylab.agents.sac.trainer import sac_config
from raylab.policy.model_based.training_mixin import TrainingSpec

from .policy import DynaSACTorchPolicy


@trainer.configure
@trainer.option("module/type", default="ModelBasedSAC")
@trainer.option("torch_optimizer/models", default={"type": "Adam", "lr": 1e-3})
@trainer.option(
    "model_training", default=TrainingSpec().to_dict(), help=TrainingSpec.__doc__
)
@trainer.option(
    "model_warmup",
    default=TrainingSpec().to_dict(),
    help="""Specifications for model warm-up.

    Same configurations as 'model_training'.
    """,
)
@trainer.option("exploration_config/pure_exploration_steps", default=1000)
@trainer.option("evaluation_config/explore", default=False, override=True)
@sac_config
class DynaSACTrainer(ModelBasedTrainer):
    """Single agent trainer for Dyna-SAC."""

    _name = "Dyna-SAC"
    _policy = DynaSACTorchPolicy



# pylint:disable=missing-module-docstring
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import List
from typing import Optional

import torch
import torch.nn as nn
from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box

import raylab.pytorch.nn as nnx
from raylab.pytorch.nn.init import initialize_
from raylab.pytorch.nn.utils import get_activation


@dataclass
class StateMLPSpec(DataClassJsonMixin):
    """Specifications for creating a multilayer perceptron.

    Args:
        units: Number of units in each hidden layer
        activation: Nonlinearity following each linear layer
        layer_norm: Whether to apply layer normalization between each linear layer
            and following activation
    """

    units: List[int] = field(default_factory=list)
    activation: Optional[str] = None
    layer_norm: bool = False


class StateMLP(nnx.FullyConnected):
    """Multilayer perceptron for encoding state inputs."""

    spec_cls = StateMLPSpec

    def __init__(self, obs_space: Box, spec: StateMLPSpec):
        obs_size = obs_space.shape[0]
        super().__init__(
            obs_size, spec.units, spec.activation, layer_norm=spec.layer_norm,
        )
        self.spec = spec

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all Linear models in the encoder.

        Uses `raylab.pytorch.nn.init.initialize_` to create an initializer
        function.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        initializer = initialize_(activation=self.spec.activation, **initializer_spec)
        self.apply(initializer)


@dataclass
class StateActionMLPSpec(DataClassJsonMixin):
    """Specifications for building an MLP with state and action inputs.

    Args:
        units: Number of units in each hidden layer
        activation: Nonlinearity following each linear layer
        delay_action: Whether to apply an initial preprocessing layer on the
            observation before concatenating the action to the input.
    """

    units: List[int] = field(default_factory=list)
    activation: Optional[str] = None
    delay_action: bool = False


class StateActionMLP(nnx.StateActionEncoder):
    """Multilayer perceptron for encoding state-action inputs."""

    spec_cls = StateActionMLPSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: StateActionMLPSpec):
        obs_size = obs_space.shape[0]
        action_size = action_space.shape[0]

        super().__init__(
            obs_size,
            action_size,
            units=spec.units,
            activation=spec.activation,
            delay_action=spec.delay_action,
        )
        self.spec = spec

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all Linear models in the encoder.

        Uses `raylab.pytorch.nn.init.initialize_` to create an initializer
        function.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        initializer = initialize_(activation=self.spec.activation, **initializer_spec)
        self.apply(initializer)


class MLP(nn.Module):
    """A general purpose Multi-Layer Perceptron."""

    def __init__(
        self,
        in_features,
        out_features,
        hidden_features,
        state_features=None,
        num_blocks=2,
        activation="ReLU",
        activate_output=False,
    ):
        # pylint:disable=too-many-arguments
        super().__init__()
        activation = get_activation(activation)
        self.stateful = bool(state_features)
        if self.stateful:
            self.initial_layer = nn.Linear(
                in_features + state_features, hidden_features
            )
        else:
            self.initial_layer = nn.Linear(in_features, hidden_features)

        layers = [activation()]
        layers += [
            layer
            for _ in range(num_blocks)
            for layer in (nn.Linear(hidden_features, hidden_features), activation())
        ]
        layers += [nn.Linear(hidden_features, out_features)]

        if activate_output:
            layers += [activation()]

        self.sequential = nn.Sequential(*layers)

    def forward(self, inputs, params: Optional[Dict[str, torch.Tensor]] = None):
        # pylint:disable=arguments-differ
        if self.stateful:
            if params is None:
                raise ValueError("Parameters required for stateful mlp.")
            out = self.initial_layer(torch.cat([inputs, params["state"]], dim=-1))
        else:
            out = self.initial_layer(inputs)

        return self.sequential(out)



"""Stochastic Actor with Normalizing Flows density approximation."""
import warnings

import gym.spaces as spaces
import torch
import torch.nn as nn
from ray.rllib.utils import override

import raylab.policy.modules.networks as networks
import raylab.pytorch.nn as nnx
import raylab.pytorch.nn.distributions as ptd
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge

from .stochastic_actor_mixin import StochasticPolicy


BASE_CONFIG = {
    "conditional_prior": True,
    "obs_encoder": {"units": (64, 64), "activation": "ReLU"},
    "initializer_options": {},
    "num_flows": 4,
    "conditional_flow": False,
    "flow": {
        "type": "AffineCouplingTransform",
        "transform_net": {"type": "MLP", "num_blocks": 0},
    },
}


class NormalizingFlowActorMixin:
    """Stochastic actor module with Normalizing Flow density estimator."""

    # pylint:disable=too-few-public-methods

    def _make_actor(self, obs_space, action_space, config):
        config = deep_merge(
            BASE_CONFIG,
            config.get("actor", {}),
            False,
            ["obs_encoder", "flow", "initalizer_options"],
            ["flow"],
        )
        assert isinstance(
            action_space, spaces.Box
        ), f"Normalizing Flow incompatible with action space type {type(action_space)}"

        # PRIOR ========================================================================
        params_module, base_dist = self._make_actor_prior(
            obs_space, action_space, config
        )
        # NormalizingFlow ==============================================================
        transforms = self._make_actor_transforms(
            action_space, params_module.state_size, config
        )
        dist_module = ptd.TransformedDistribution(
            base_dist=base_dist, transform=ptd.flows.CompositeTransform(transforms),
        )

        return {"actor": StochasticPolicy(params_module, dist_module)}

    @staticmethod
    def _make_actor_prior(obs_space, action_space, config):
        # Ensure we're not encoding the observation for nothing
        if config["conditional_prior"] or config["conditional_flow"]:
            obs_encoder = nnx.FullyConnected(
                obs_space.shape[0], **config["obs_encoder"]
            )
        else:
            warnings.warn("Policy is blind to the observations")
            obs_encoder = nnx.FullyConnected(obs_space.shape[0], units=())

        obs_encoder.apply(
            initialize_(
                activation=config["obs_encoder"].get("activation"),
                **config["initializer_options"],
            )
        )

        params_module = NFNormalParams(obs_encoder, action_space, config)
        base_dist = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)
        return params_module, base_dist

    @staticmethod
    def _make_actor_transforms(action_space, state_size, config):
        act_size = action_space.shape[0]
        flow_config = config["flow"].copy()
        cls = getattr(ptd.flows, flow_config.pop("type"))

        if issubclass(cls, ptd.flows.CouplingTransform):
            net_config = flow_config.pop("transform_net")
            net_config.setdefault("hidden_features", act_size)
            transform_net = getattr(networks, net_config.pop("type"))

            def transform_net_create_fn(in_features, out_features):
                return transform_net(
                    in_features,
                    out_features,
                    state_features=state_size if config["conditional_flow"] else None,
                    **net_config,
                )

            masks = [
                ptd.flows.masks.create_alternating_binary_mask(act_size, bool(i % 2))
                for i in range(config["num_flows"])
            ]
            transforms = [cls(m, transform_net_create_fn, **flow_config) for m in masks]

        else:
            raise NotImplementedError(f"Unsupported flow type {cls}")

        squash = ptd.flows.TanhSquashTransform(
            low=torch.as_tensor(action_space.low),
            high=torch.as_tensor(action_space.high),
            event_dim=1,
        )
        return transforms + [squash]


class NFNormalParams(nn.Module):
    """Maps inputs to distribution parameters for Normalizing Flows."""

    def __init__(self, obs_encoder, action_space, config):
        super().__init__()
        self.obs_encoder = obs_encoder
        self.state_size = self.obs_encoder.out_features

        act_size = action_space.shape[0]
        if config["conditional_prior"]:
            self.params = nnx.NormalParams(self.state_size, act_size)
        else:
            self.params = nnx.StdNormalParams(1, act_size)

    @override(nn.Module)
    def forward(self, obs):  # pylint:disable=arguments-differ
        state = self.obs_encoder(obs)
        params = self.params(state)
        # For use in conditional flows later
        params["state"] = state
        return params



"""Policy for Dyna-SAC in PyTorch."""
from raylab.agents.sac import SACTorchPolicy
from raylab.policy.losses import DynaSoftCDQLearning
from raylab.policy.losses import ModelEnsembleMLE
from raylab.policy.model_based import EnvFnMixin
from raylab.policy.model_based import ModelTrainingMixin
from raylab.pytorch.optim.utils import build_optimizer


class DynaSACTorchPolicy(ModelTrainingMixin, EnvFnMixin, SACTorchPolicy):
    """Model-based policy por Dyna-SAC."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_model_loss()

    @property
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.sac.dyna import DynaSACTrainer

        return DynaSACTrainer.options

    @property
    def model_training_loss(self):
        return self.loss_model

    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        optimizers.update(
            models=build_optimizer(
                self.module.models, config=self.config["torch_optimizer"]["models"]
            )
        )
        return optimizers

    def _setup_model_loss(self):
        self.loss_model = ModelEnsembleMLE(self.module.models)

    def _setup_critic_loss(self):
        self.loss_critic = DynaSoftCDQLearning(
            self.module.critics,
            self.module.models,
            self.module.target_critics,
            self.module.actor,
        )
        self.loss_critic.gamma = self.config["gamma"]
        self.loss_critic.seed(self.config["seed"])

    def _set_reward_hook(self):
        self.loss_critic.set_reward_fn(self.reward_fn)

    def _set_termination_hook(self):
        self.loss_critic.set_termination_fn(self.termination_fn)



"""Model-based architecture with disjoint model, actor, and critic."""

from .abstract import AbstractModelActorCritic
from .mixins import StateValueMixin
from .mixins import StochasticActorMixin
from .mixins import StochasticModelMixin


# pylint:disable=abstract-method
class SimpleModelBased(
    StochasticModelMixin,
    StochasticActorMixin,
    StateValueMixin,
    AbstractModelActorCritic,
):
    """Module architecture with stochastic actor and model, and state value function."""



"""Policy for MAGE using PyTorch."""
from raylab.agents.sop import SOPTorchPolicy
from raylab.policy import EnvFnMixin
from raylab.policy import ModelTrainingMixin
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import MAGE
from raylab.policy.losses import ModelEnsembleMLE
from raylab.policy.losses.mage import MAGEModules
from raylab.pytorch.optim import build_optimizer


class MAGETorchPolicy(ModelTrainingMixin, EnvFnMixin, SOPTorchPolicy):
    """MAGE policy in PyTorch to use with RLlib.

    Attributes:
        loss_model: maximum likelihood loss for model ensemble
        loss_actor: deterministic policy gradient loss
        loss_critic: model-based action-value-gradient estimator loss
    """

    dist_class = WrapDeterministicPolicy

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)

        module = self.module
        self.loss_model = ModelEnsembleMLE(module.models)
        mage_modules = MAGEModules(
            critics=module.critics,
            target_critics=module.target_critics,
            policy=module.actor,
            target_policy=module.target_actor,
            models=module.models,
        )
        self.loss_critic = MAGE(mage_modules)
        self.loss_critic.gamma = self.config["gamma"]
        self.loss_critic.lambda_ = self.config["lambda"]

    @property
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.mage import MAGETrainer

        return MAGETrainer.options

    @property
    def model_training_loss(self):
        return self.loss_model

    def compile(self):
        super().compile()
        for loss in (self.loss_model, self.loss_actor, self.loss_critic):
            loss.compile()

    def _set_reward_hook(self):
        self.loss_critic.set_reward_fn(self.reward_fn)

    def _set_termination_hook(self):
        self.loss_critic.set_termination_fn(self.termination_fn)

    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        optimizers["models"] = build_optimizer(
            self.module.models, self.config["torch_optimizer"]["models"]
        )
        return optimizers



"""Trainer and configuration for ACKTR."""
from raylab.agents import trainer
from raylab.agents.trpo import TRPOTrainer

from .policy import ACKTRTorchPolicy
from .policy import DEFAULT_OPTIM_CONFIG


@trainer.configure
@trainer.option("torch_optimizer", DEFAULT_OPTIM_CONFIG, override=True)
class ACKTRTrainer(TRPOTrainer):
    """Single agent trainer for ACKTR."""

    _name = "ACKTR"
    _policy = ACKTRTorchPolicy



"""
MIT License

Copyright (c) 2019 Conor Durkan, Artur Bekasov, Iain Murray, George Papamakarios

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Implementations of various coupling layers.

Slightly modified from:
https://github.com/bayesiains/nsf/blob/master/nde/transforms/coupling.py
"""
import warnings
from abc import ABCMeta
from abc import abstractmethod
from typing import Dict

import numpy as np
import torch
from ray.rllib.utils import override

from .abstract import ConditionalTransform
from .nonlinearities import PiecewiseRationalQuadraticCDF
from .splines import DEFAULT_MIN_BIN_HEIGHT
from .splines import DEFAULT_MIN_BIN_WIDTH
from .splines import DEFAULT_MIN_DERIVATIVE
from .splines import unconstrained_rational_quadratic_spline
from .utils import sum_rightmost


class CouplingTransform(ConditionalTransform, metaclass=ABCMeta):
    """A base class for coupling layers. Supports 1D inputs (*, D), where D >= 2."""

    def __init__(
        self,
        mask,
        transform_net_create_fn,
        unconditional_transform=None,
        *,
        event_dim=1,
    ):
        """
        Constructor.
        Args:
            mask: a 1-dim tensor, tuple or list. It indexes inputs as follows:
                * If `mask[i] > 0`, `input[i]` will be transformed.
                * If `mask[i] <= 0`, `input[i]` will be passed unchanged.
        """
        mask = torch.as_tensor(mask)
        if mask.dim() != 1:
            raise ValueError("Mask must be a 1-dim tensor.")
        if mask.numel() <= 1:
            raise ValueError("Coupling dimension must greater than 1.")

        super().__init__(event_dim=event_dim)

        self.features = len(mask)
        indexes = torch.arange(self.features)
        self.register_buffer("identity_features", indexes.masked_select(mask <= 0))
        self.register_buffer("transform_features", indexes.masked_select(mask > 0))
        # Number of features that condition the transform.
        self.num_identity_features = len(self.identity_features)
        # Number of features that are conditionally transformed.
        self.num_transform_features = len(self.transform_features)
        assert self.num_identity_features + self.num_transform_features == self.features

        self.transform_net = transform_net_create_fn(
            self.num_identity_features,
            self.num_transform_features * self.transform_dim_multiplier,
        )
        if unconditional_transform is None:
            self.unconditional_transform = None
        else:
            self.unconditional_transform = unconditional_transform(
                features=self.num_identity_features
            )

    @override(ConditionalTransform)
    def encode(self, inputs, params: Dict[str, torch.Tensor]):
        identity_split = inputs.index_select(-1, self.identity_features)
        transform_split = inputs.index_select(-1, self.transform_features)

        transform_params = self.transform_net(identity_split, params)
        transform_split, logabsdet = self.coupling_transform_forward(
            inputs=transform_split, transform_params=transform_params
        )

        if self.unconditional_transform is not None:
            identity_split, logabsdet_identity = self.unconditional_transform(
                identity_split, params
            )
            logabsdet += logabsdet_identity

        outputs = torch.empty_like(inputs)
        outputs.index_copy_(-1, self.identity_features, identity_split)
        outputs.index_copy_(-1, self.transform_features, transform_split)

        return outputs, logabsdet

    @override(ConditionalTransform)
    def decode(self, inputs, params: Dict[str, torch.Tensor]):
        identity_split = inputs.index_select(-1, self.identity_features)
        transform_split = inputs.index_select(-1, self.transform_features)

        logabsdet = 0.0
        if self.unconditional_transform is not None:
            identity_split, logabsdet = self.unconditional_transform.inverse(
                identity_split, params
            )

        transform_params = self.transform_net(identity_split, params)
        transform_split, logabsdet_split = self.coupling_transform_inverse(
            inputs=transform_split, transform_params=transform_params
        )
        logabsdet += logabsdet_split

        outputs = torch.empty_like(inputs)
        outputs.index_copy_(-1, self.identity_features, identity_split)
        outputs.index_copy_(-1, self.transform_features, transform_split)

        return outputs, logabsdet

    @property
    @abstractmethod
    def transform_dim_multiplier(self):
        """Number of features to output for each transform dimension."""

    @abstractmethod
    def coupling_transform_forward(self, inputs, transform_params):
        """Forward pass of the coupling transform."""

    @abstractmethod
    def coupling_transform_inverse(self, inputs, transform_params):
        """Inverse of the coupling transform."""


class AffineCouplingTransform(CouplingTransform):
    """An affine coupling layer that scales and shifts part of the variables.
    Reference:
    > L. Dinh et al., Density estimation using Real NVP, ICLR 2017.
    """

    @property
    @override(CouplingTransform)
    def transform_dim_multiplier(self):
        return 2

    def _scale_and_shift(self, transform_params):
        unconstrained_scale = transform_params[..., self.num_transform_features :]
        shift = transform_params[..., : self.num_transform_features]

        # Obtain Scale
        # Option 1: use exp()
        log_scale = unconstrained_scale * 0.01
        scale = log_scale.exp()

        # Option 2: use softplus()
        # scale = (F.softplus(unconstrained_scale) + 1e-3).clamp(0, 3)
        # log_scale = scale.log()

        # Option 3: use sigmoid()
        # scale = torch.sigmoid(unconstrained_scale + 2) + 1e-3
        # log_scale = scale.log()

        return scale, shift, log_scale

    @override(CouplingTransform)
    def coupling_transform_forward(self, inputs, transform_params):
        scale, shift, log_scale = self._scale_and_shift(transform_params)
        outputs = inputs * scale + shift
        logabsdet = sum_rightmost(log_scale, self.event_dim)
        return outputs, logabsdet

    @override(CouplingTransform)
    def coupling_transform_inverse(self, inputs, transform_params):
        scale, shift, log_scale = self._scale_and_shift(transform_params)
        outputs = (inputs - shift) / scale
        logabsdet = -sum_rightmost(log_scale, self.event_dim)
        return outputs, logabsdet


class AdditiveCouplingTransform(AffineCouplingTransform):
    """An additive coupling layer, i.e. an affine coupling layer without scaling.
    Reference:
    > L. Dinh et al., NICE:  Non-linear  Independent  Components  Estimation,
    > arXiv:1410.8516, 2014.
    """

    @property
    @override(CouplingTransform)
    def transform_dim_multiplier(self):
        return 1

    @override(AffineCouplingTransform)
    def _scale_and_shift(self, transform_params):
        shift = transform_params
        scale = torch.ones_like(shift)
        return scale, shift, torch.zeros_like(scale)


class PiecewiseCouplingTransform(CouplingTransform):
    """
    Base class for piecewise coupling transforms, which define their inverse externally.
    """

    @override(CouplingTransform)
    def coupling_transform_forward(self, inputs, transform_params):
        return self._coupling_transform(inputs, transform_params, reverse=False)

    @override(CouplingTransform)
    def coupling_transform_inverse(self, inputs, transform_params):
        return self._coupling_transform(inputs, transform_params, reverse=True)

    def _coupling_transform(self, inputs, transform_params, reverse: bool = False):

        # For batched 1D data, reshape transform_params from (*, D*?) to (*, D, ?)
        transform_params = torch.reshape(transform_params, inputs.shape + (-1,))

        outputs, logabsdet = self._piecewise_cdf(
            inputs, transform_params, reverse=reverse
        )

        return outputs, sum_rightmost(logabsdet, self.event_dim)

    @abstractmethod
    def _piecewise_cdf(self, inputs, transform_params, reverse: bool = False):
        pass


class PiecewiseRQSCouplingTransform(PiecewiseCouplingTransform):
    """Piecewise Rational Quadratic Spline Coupling Transform."""

    def __init__(
        self,
        mask,
        transform_net_create_fn,
        num_bins=10,
        tail_bound=1.0,
        apply_unconditional_transform=False,
        img_shape=None,
        min_bin_width=DEFAULT_MIN_BIN_WIDTH,
        min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
        min_derivative=DEFAULT_MIN_DERIVATIVE,
        **kwargs,
    ):
        # pylint:disable=too-many-arguments
        self.num_bins = num_bins
        self.min_bin_width = min_bin_width
        self.min_bin_height = min_bin_height
        self.min_derivative = min_derivative
        self.tail_bound = tail_bound

        if apply_unconditional_transform:

            def unconditional_transform(features):
                return PiecewiseRationalQuadraticCDF(
                    shape=[features] + (img_shape if img_shape else []),
                    num_bins=num_bins,
                    tail_bound=tail_bound,
                    min_bin_width=min_bin_width,
                    min_bin_height=min_bin_height,
                    min_derivative=min_derivative,
                )

        else:
            unconditional_transform = None

        super().__init__(
            mask,
            transform_net_create_fn,
            unconditional_transform=unconditional_transform,
            **kwargs,
        )

        if hasattr(self.transform_net, "hidden_features"):
            self.scale = 1.0 / np.sqrt(self.transform_net.hidden_features)
        else:
            self.scale = 1.0
            warnings.warn(
                "Inputs to the softmax are not scaled down: init might be bad."
            )

    @property
    @override(CouplingTransform)
    def transform_dim_multiplier(self):
        # Always use linear tails
        return self.num_bins * 3 - 1

    def _piecewise_cdf(self, inputs, transform_params, reverse: bool = False):
        unnormalized_widths = transform_params[..., : self.num_bins]
        unnormalized_heights = transform_params[..., self.num_bins : 2 * self.num_bins]
        unnormalized_derivatives = transform_params[..., 2 * self.num_bins :]

        unnormalized_widths *= self.scale
        unnormalized_heights *= self.scale

        # Always use linear tails
        return unconstrained_rational_quadratic_spline(
            inputs=inputs,
            unnormalized_widths=unnormalized_widths,
            unnormalized_heights=unnormalized_heights,
            unnormalized_derivatives=unnormalized_derivatives,
            inverse=reverse,
            tail_bound=self.tail_bound,
            min_bin_width=self.min_bin_width,
            min_bin_height=self.min_bin_height,
            min_derivative=self.min_derivative,
        )



"""Utilities for module initialization."""
import functools
import inspect
from typing import Callable
from typing import Optional
from typing import Union

import torch.nn as nn
from torch import Tensor


def get_initializer(name: Optional[str]) -> Callable[[Tensor], None]:
    """Return initializer function given its name.

    Arguments:
        name: The initializer function's name. If None, returns a no-op callable
    """
    if name is None:
        return lambda _: None

    name_ = name + "_"
    if name in dir(nn.init) and name_ in dir(nn.init):
        func = getattr(nn.init, name_)
        return func
    raise ValueError(f"Couldn't find initializer with name '{name}'")


NONLINEARITY_MAP = {
    "Sigmoid": "sigmoid",
    "Tanh": "tanh",
    "ReLU": "relu",
    "ELU": "relu",
    "LeakyReLU": "leaky_relu",
}


def initialize_(
    name: Optional[str] = None, activation: Union[str, dict] = None, **options
) -> Callable[[nn.Module], None]:
    """Return a callable to apply an initializer with the given name and options.

    If `gain` is part of the initializer's argspec and is not specified in options,
    the recommended value from `torch.nn.init.calculate_gain` is used.

    Arguments:
        name: Initializer function name
        activation: Optional specification of the activation function that
            follows linear layers
        **options: Keyword arguments to pass to the initializer

    Returns:
        A callable to be used with `nn.Module.apply`.
    """

    initializer = get_initializer(name)

    if isinstance(activation, dict):
        activation = activation["name"]
        options.update(activation.get("options", {}))

    if (
        activation in NONLINEARITY_MAP
        and "gain" not in options
        and "gain" in inspect.signature(initializer).parameters
    ):
        recommended_gain = nn.init.calculate_gain(
            NONLINEARITY_MAP[activation], param=options.get("negative_slope")
        )
        options["gain"] = recommended_gain
    func_ = functools.partial(initializer, **options)

    def init(module):
        if isinstance(module, nn.Linear):
            func_(module.weight)
            if module.bias is not None:
                nn.init.constant_(module.bias, 0)

    return init



"""Trainer and configuration for SVG(1) with maximum entropy."""
from ray.rllib import SampleBatch
from ray.rllib.utils import override

from raylab.agents import trainer
from raylab.agents.model_based import set_policy_with_env_fn
from raylab.agents.off_policy import OffPolicyTrainer
from raylab.utils.replay_buffer import ReplayField

from .policy import SoftSVGTorchPolicy


TORCH_OPTIMIZERS = {
    "model": {"type": "Adam", "lr": 1e-3},
    "actor": {"type": "Adam", "lr": 1e-3},
    "critic": {"type": "Adam", "lr": 1e-3},
    "alpha": {"type": "Adam", "lr": 1e-3},
}

DEFAULT_MODULE = {"type": "MaxEntModelBased-v0", "critic": {"target_vf": True}}
EXPLORATION_CONFIG = {
    "type": "raylab.utils.exploration.StochasticActor",
    "pure_exploration_steps": 1000,
}


@trainer.configure
@trainer.option(
    "target_entropy",
    None,
    help="""
Target entropy to optimize the temperature parameter towards
If "auto", will use the heuristic provided in the SAC paper,
H = -dim(A), where A is the action space
""",
)
@trainer.option("torch_optimizer", TORCH_OPTIMIZERS, override=True)
@trainer.option(
    "vf_loss_coeff",
    1.0,
    help="Weight of the fitted V loss in the joint model-value loss",
)
@trainer.option(
    "max_is_ratio", 5.0, help="Clip importance sampling weights by this value"
)
@trainer.option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@trainer.option("module", DEFAULT_MODULE, override=True)
@trainer.option("exploration_config", EXPLORATION_CONFIG, override=True)
@trainer.option("evaluation_config/explore", False, override=True)
class SoftSVGTrainer(OffPolicyTrainer):
    """Single agent trainer for SoftSVG."""

    # pylint:disable=attribute-defined-outside-init
    _name = "SoftSVG"
    _policy = SoftSVGTorchPolicy

    @override(OffPolicyTrainer)
    def _init(self, config, env_creator):
        super()._init(config, env_creator)
        set_policy_with_env_fn(self.workers, fn_type="reward")

    @override(OffPolicyTrainer)
    def build_replay_buffer(self, config):
        super().build_replay_buffer(config)
        self.replay.add_fields(ReplayField(SampleBatch.ACTION_LOGP))



"""Support for modules with stochastic policies."""
from typing import List

import torch
import torch.nn as nn
from ray.rllib.utils import override

import raylab.pytorch.nn as nnx
import raylab.pytorch.nn.distributions as ptd
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge


BASE_CONFIG = {
    "residual": True,
    "input_dependent_scale": False,
    # Number of independent models. If 0 return a single model accessible via the
    # 'model' atribute. If greater than 0, return a nn.ModuleList of models accessible
    # via the 'models' attribute.
    "ensemble_size": 0,
    "encoder": {"units": (32, 32), "activation": "ReLU", "delay_action": True},
    "initializer_options": {"name": "xavier_uniform"},
}


class StochasticModelMixin:
    """Adds constructor for modules with stochastic dynamics model."""

    # pylint:disable=too-few-public-methods

    def _make_model(self, obs_space, action_space, config):
        config = self.process_config(config)
        return self.build_models(obs_space, action_space, config)

    @staticmethod
    def process_config(config):
        """Fill in default configuration for models."""
        return deep_merge(BASE_CONFIG, config.get("model", {}), False, ["encoder"])

    def build_models(self, obs_space, action_space, config):
        """Decide whether to return a single model or an ensemble."""
        if config["ensemble_size"] == 0:
            return {"model": self.build_single_model(obs_space, action_space, config)}

        models = [
            self.build_single_model(obs_space, action_space, config)
            for _ in range(config["ensemble_size"])
        ]
        return {"models": nn.ModuleList(models)}

    @staticmethod
    def build_single_model(obs_space, action_space, config):
        """Build a stochastic model module."""
        params_module = GaussianDynamicsParams(obs_space, action_space, config)
        dist_module = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)

        return StochasticModel.assemble(params_module, dist_module, config)


class GaussianDynamicsParams(nn.Module):
    """
    Neural network module mapping inputs to Normal distribution parameters.
    """

    def __init__(self, obs_space, action_space, config):
        super().__init__()
        obs_size, act_size = obs_space.shape[0], action_space.shape[0]
        self.logits = nnx.StateActionEncoder(obs_size, act_size, **config["encoder"])
        self.logits.apply(
            initialize_(
                activation=config["encoder"].get("activation"),
                **config["initializer_options"]
            )
        )
        self.params = nnx.NormalParams(
            self.logits.out_features,
            obs_size,
            input_dependent_scale=config["input_dependent_scale"],
        )

    @override(nn.Module)
    def forward(self, obs, actions):  # pylint:disable=arguments-differ
        return self.params(self.logits(obs, actions))


class StochasticModel(nn.Module):
    """Represents a stochastic model as a conditional distribution module."""

    def __init__(self, params_module, dist_module):
        super().__init__()
        self.params = params_module
        self.dist = dist_module

    @override(nn.Module)
    def forward(self, obs, action):  # pylint:disable=arguments-differ
        return self.params(obs, action)

    @torch.jit.export
    def sample(self, obs, action, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        params = self(obs, action)
        return self.dist.sample(params, sample_shape)

    @torch.jit.export
    def rsample(self, obs, action, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        params = self(obs, action)
        return self.dist.rsample(params, sample_shape)

    @torch.jit.export
    def log_prob(self, obs, action, next_obs):
        """
        Returns the log probability density/mass function evaluated at `next_obs`.
        """
        params = self(obs, action)
        return self.dist.log_prob(next_obs, params)

    @torch.jit.export
    def cdf(self, obs, action, next_obs):
        """Returns the cumulative density/mass function evaluated at `next_obs`."""
        params = self(obs, action)
        return self.dist.cdf(next_obs, params)

    @torch.jit.export
    def icdf(self, obs, action, prob):
        """Returns the inverse cumulative density/mass function evaluated at `prob`."""
        params = self(obs, action)
        return self.dist.icdf(prob, params)

    @torch.jit.export
    def entropy(self, obs, action):
        """Returns entropy of distribution."""
        params = self(obs, action)
        return self.dist.entropy(params)

    @torch.jit.export
    def perplexity(self, obs, action):
        """Returns perplexity of distribution."""
        params = self(obs, action)
        return self.dist.perplexity(params)

    @torch.jit.export
    def reproduce(self, obs, action, next_obs):
        """Produce a reparametrized sample with the same value as `next_obs`."""
        params = self(obs, action)
        return self.dist.reproduce(next_obs, params)

    @classmethod
    def assemble(cls, params_module, dist_module, config):
        """Return a residual or normal stochastic model depending on configuration."""
        if config["residual"]:
            return ResidualStochasticModel(params_module, dist_module)
        return cls(params_module, dist_module)


class ResidualStochasticModel(StochasticModel):
    """
    Represents a stochastic model as a conditional distribution module that predicts
    residuals.
    """

    @override(StochasticModel)
    @torch.jit.export
    def sample(self, obs, action, sample_shape: List[int] = ()):
        params = self(obs, action)
        res, log_prob = self.dist.sample(params, sample_shape)
        return obs + res, log_prob

    @override(StochasticModel)
    @torch.jit.export
    def rsample(self, obs, action, sample_shape: List[int] = ()):
        params = self(obs, action)
        res, log_prob = self.dist.rsample(params, sample_shape)
        return obs + res, log_prob

    @override(StochasticModel)
    @torch.jit.export
    def log_prob(self, obs, action, next_obs):
        params = self(obs, action)
        return self.dist.log_prob(next_obs - obs, params)

    @override(StochasticModel)
    @torch.jit.export
    def cdf(self, obs, action, next_obs):
        params = self(obs, action)
        return self.dist.cdf(next_obs - obs, params)

    @override(StochasticModel)
    @torch.jit.export
    def icdf(self, obs, action, prob):
        params = self(obs, action)
        residual = self.dist.icdf(prob, params)
        return obs + residual

    @override(StochasticModel)
    @torch.jit.export
    def reproduce(self, obs, action, next_obs):
        params = self(obs, action)
        sample_, log_prob_ = self.dist.reproduce(next_obs - obs, params)
        return obs + sample_, log_prob_



"""Loss functions for Maximum Likelihood Estimation."""
from typing import Tuple

import torch
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.policy.modules.model.stochastic.ensemble import StochasticModelEnsemble
from raylab.policy.modules.model.stochastic.single import StochasticModel
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict
from raylab.utils.dictionaries import get_keys

from .abstract import Loss


class MaximumLikelihood(Loss):
    """Loss function for model learning of single transitions.

    Args:
        model: parametric stochastic model
    """

    batch_keys: Tuple[str, str, str] = (
        SampleBatch.CUR_OBS,
        SampleBatch.ACTIONS,
        SampleBatch.NEXT_OBS,
    )

    def __init__(self, model: StochasticModel):
        self.model = model

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute Maximum Likelihood Estimation (MLE) model loss.

        Returns:
            A tuple containg a 0d loss tensor and a dictionary of loss
            statistics
        """
        obs, actions, next_obs = get_keys(batch, *self.batch_keys)
        loss = -self.model_likelihood(obs, actions, next_obs).mean()
        return loss, {"loss(model)": loss.item()}

    def model_likelihood(
        self, obs: Tensor, actions: Tensor, next_obs: Tensor
    ) -> Tensor:
        """Compute likelihood of a transition under the model."""
        return self.model.log_prob(obs, actions, next_obs)


class ModelEnsembleMLE(Loss):
    """MLE loss function for ensemble of models.

    Args:
        models: the list of models
    """

    batch_keys: Tuple[str, str, str] = MaximumLikelihood.batch_keys

    def __init__(self, models: StochasticModelEnsemble):
        self.models = models

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute Maximum Likelihood Estimation (MLE) loss for each model.

        Returns:
            A tuple with a 1d loss tensor containing each model's loss and a
            dictionary of loss statistics
        """
        obs, actions, next_obs = map(
            self.expand_foreach_model, get_keys(batch, *self.batch_keys)
        )
        loss = -self.models.log_prob(obs, actions, next_obs).mean(dim=-1)
        info = {f"loss(models[{i}])": s for i, s in enumerate(loss.tolist())}
        return loss, info

    def expand_foreach_model(self, tensor: Tensor) -> Tensor:
        """Add first dimension to tensor with the size of the model ensemble.

        Args:
            tensor: Tensor of shape `S`

        Returns:
            Tensor `tensor` expanded to shape `(N,) + S`
        """
        return tensor.expand((len(self.models),) + tensor.shape)

    def compile(self):
        self.models = torch.jit.script(self.models)



"""Mixins for model-based policies."""

from .envfn_mixin import EnvFnMixin
from .sampling_mixin import ModelSamplingMixin
from .training_mixin import ModelTrainingMixin

__all__ = [
    "ModelTrainingMixin",
    "ModelSamplingMixin",
    "EnvFnMixin",
]



# pylint:disable=missing-module-docstring
from raylab.agents.mapo import MAPOTorchPolicy
from raylab.policy.losses import ModelEnsembleMLE
from raylab.policy.losses import SPAML


class MlMAPOTorchPolicy(MAPOTorchPolicy):
    """PyTorch policy for MAPO-MLE.

    Effectively substitutes a Maximum Likelihood loss for the policy-aware model
    loss in MAPO.
    """

    # pylint:disable=too-many-ancestors

    @property
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.mapo.mle import MlMAPOTrainer

        return MlMAPOTrainer.options

    @property
    def model_training_loss(self):
        return self.loss_mle

    @property
    def model_warmup_loss(self):
        return self.loss_mle

    def _setup_model_loss(self):
        # Placeholder
        self.loss_paml = SPAML(
            self.module.models, self.module.actor, self.module.critics
        )

        self.loss_mle = ModelEnsembleMLE(self.module.models)

    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    def _set_termination_hook(self):
        self.loss_actor.set_termination_fn(self.termination_fn)



"""NN Architecture with disjoint model, actor, and critic."""
from raylab.utils.dictionaries import deep_merge

from .abstract import AbstractModelActorCritic
from .mixins import ActionValueMixin
from .mixins import DeterministicActorMixin
from .mixins import StochasticModelMixin


BASE_CONFIG = {
    "actor": {
        "beta": 1.2,
        "smooth_target_policy": False,
        "target_gaussian_sigma": 0.3,
        "perturbed_policy": False,
        "encoder": {
            "units": (32, 32),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
            "layer_norm": False,
        },
    },
    "critic": {
        "double_q": False,
        "encoder": {
            "units": (32, 32),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
            "delay_action": True,
        },
    },
    "model": {
        "residual": False,
        "input_dependent_scale": False,
        "encoder": {
            "units": (32, 32),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
            "delay_action": True,
        },
    },
}


class ModelBasedDDPG(
    StochasticModelMixin,
    DeterministicActorMixin,
    ActionValueMixin,
    AbstractModelActorCritic,
):
    """Module architecture used in MAGE."""

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config, False, ["actor", "critic", "model"])
        super().__init__(obs_space, action_space, config)



"""Environment model handling mixins for TorchPolicy."""
import collections
import copy
import itertools
import time
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
from dataclasses_json import DataClassJsonMixin
from ray.rllib import SampleBatch
from torch.utils.data import DataLoader
from torch.utils.data import RandomSampler

from raylab.policy.losses.abstract import Loss
from raylab.pytorch.utils import TensorDictDataset
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict


@dataclass
class DataloaderSpec(DataClassJsonMixin):
    """Specifications for creating the data loader.

    Attributes:
        batch_size: Size of minibatch for dynamics model training
        replacement: Whether to sample transitions with replacement
    """

    batch_size: int = 256
    replacement: bool = False

    def __post_init__(self):
        assert self.batch_size > 0, "Model batch size must be positive"


@dataclass
class TrainingSpec(DataClassJsonMixin):
    """Specifications for training the model.

    Attributes:
        dataloader: specifications for creating the data loader
        max_epochs: Maximum number of full model passes through the data
        max_grad_steps: Maximum number of model gradient steps
        improvement_threshold: Minimum expected relative improvement in model
            validation loss
        patience_epochs: Number of epochs to wait for any of the models to
            improve on the validation dataset before early stopping
        max_time: Maximum time in seconds for training the model. We
            check this after each epoch (not minibatch)
    """

    dataloader: DataloaderSpec = field(default_factory=DataloaderSpec, repr=True)
    max_epochs: Optional[int] = 120
    max_grad_steps: Optional[int] = 120
    max_time: Optional[float] = 20
    patience_epochs: Optional[int] = 5
    improvement_threshold: Optional[float] = 0.01

    def __post_init__(self):
        assert (
            not self.max_epochs or self.max_epochs > 0
        ), "Cannot train model for a negative number of epochs"
        assert not self.max_grad_steps or self.max_grad_steps > 0
        assert (
            not self.max_time or self.max_time > 0
        ), "Maximum training time must be positive"
        assert (
            not self.patience_epochs or self.patience_epochs > 0
        ), "Must wait a positive number of epochs for any model to improve"
        assert (
            self.improvement_threshold is None or self.improvement_threshold >= 0
        ), "Improvement threshold must be nonnegative"
        assert (
            self.max_epochs
            or self.max_grad_steps
            or (self.improvement_threshold is not None and self.patience_epochs)
        ), "Need at least one stopping criterion"


ModelSnapshot = collections.namedtuple("ModelSnapshot", "epoch loss state_dict")


@dataclass
class Evaluator:
    """Evaluates models and saves snapshots.

    Holds snapshots for each model. A snapshot contains the epoch in which the
    model was evaluated, its validation loss, and its state dict.

    Note:
        Upon creation, evaluates models on validation data to set initial
        snapshots.

    Args:
        models: the model ensemble
        loss_fn: the loss function for model ensemble
        improvement_threshold: Minimum expected relative improvement in model
            validation loss
        patience_epochs: Number of epochs to wait for any of the models to
            improve on the validation dataset before early stopping. If None,
            disables eary stopping.
    """

    models: nn.ModuleList
    loss_fn: Loss
    eval_tensors: TensorDict
    improvement_threshold: float
    patience_epochs: Optional[int]

    def __post_init__(self):
        eval_losses, _ = self.loss_fn(self.eval_tensors)
        eval_losses = eval_losses.tolist()
        self._snapshots = [
            ModelSnapshot(epoch=-1, loss=loss, state_dict=copy.deepcopy(m.state_dict()))
            for m, loss in zip(self.models, eval_losses)
        ]

    def validate(self, epoch: int) -> Tuple[bool, StatDict]:
        """Evaluate models on holdout data and update snapshots.

        Args:
            epoch: the epoch number

        Returns:
            A tuple with two values: whether or not to early stop training based
            on validation loss improvement and a dict with validation loss info
        """
        eval_losses, eval_info = self.loss_fn(self.eval_tensors)
        eval_losses = eval_losses.tolist()
        eval_info = {"eval_" + k: v for k, v in eval_info.items()}

        self._update_snapshots(epoch, eval_losses)

        patience_epochs = self.patience_epochs or float("inf")
        early_stop = epoch - max(s.epoch for s in self._snapshots) >= patience_epochs
        return early_stop, eval_info

    def _update_snapshots(self, epoch: int, eval_losses: List[float]):
        snapshots = self._snapshots
        threshold = self.improvement_threshold

        def updated_snapshot(model, snap, cur_loss):
            if (snap.loss - cur_loss) / snap.loss > threshold:
                return ModelSnapshot(
                    epoch=epoch,
                    loss=cur_loss,
                    state_dict=copy.deepcopy(model.state_dict()),
                )
            return snap

        self._snapshots = [
            updated_snapshot(model=m, snap=s, cur_loss=l)
            for m, s, l in zip(self.models, snapshots, eval_losses)
        ]

    def restore_models(self) -> List[float]:
        """Restore models to the best performing parameters.

        Returns:
            A list with the validation performances of each model
        """
        losses = []
        for idx, snap in enumerate(self._snapshots):
            self.models[idx].load_state_dict(snap.state_dict)
            losses += [snap.loss]
        return losses


class ModelTrainingMixin(ABC):
    """Adds model training behavior to a TorchPolicy class.

    Expects:
    * A `models` attribute in `self.module`
    * A `model_training` dict in `self.config`
    * A `model_warmup` dict in `self.config`
    * A 'models' optimizer in `self.optimizers`

    Attributes:
        model_training_spec: Specifications for training the model
        model_warmup_spec: Specifications for model warm-up
    """

    model_training_spec: TrainingSpec

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_training_spec = TrainingSpec.from_dict(self.config["model_training"])
        self.model_warmup_spec = TrainingSpec.from_dict(self.config["model_warmup"])

    @property
    @abstractmethod
    def model_training_loss(self) -> Loss:
        """Loss function used for normal model training and evaluation.

        Must return a 1d Tensor with each model's losses and an info dict.
        """

    @property
    def model_warmup_loss(self) -> Loss:
        """Loss function used for model warm-up.

        Must return a 1d Tensor with each model's losses and an info dict.
        """
        return self.model_training_loss

    def optimize_model(
        self,
        train_samples: SampleBatch,
        eval_samples: SampleBatch = None,
        warmup: bool = False,
    ) -> Tuple[List[float], StatDict]:
        """Update models with samples.

        If `spec.max_epochs` is set, training will be cut off after this many
        epochs.

        If `spec.max_grad_steps` is set, training will be cut off after this
        many model gradient steps.

        If `spec.max_time` is set, it will cut off training after this many
        seconds have elapsed.

        If `spec.improvement_threshold` is set, it will save snapshots based on
        model performance on validation data and restore models to the
        best performing parameters after training. Otherwise, it will return the
        latest models.

        If `spec.patience_epochs` is set, it will wait for at most this many
        epochs for any model to improve on the validation dataset, otherwise it
        will stop training.

        Args:
            train_samples: Training data
            eval_samples: Holdout data
            warmup: Whether to train with warm-up loss and spec

        Returns:
            A tuple with a list of each model's evaluation loss and a dictionary
            with training statistics
        """
        loss_fn = self.model_warmup_loss if warmup else self.model_training_loss
        spec = self.model_warmup_spec if warmup else self.model_training_spec

        dataloader = self._build_dataloader(train_samples, loss_fn=loss_fn, spec=spec)
        evaluator = self._setup_evaluator(eval_samples, loss_fn=loss_fn, spec=spec)

        info = self._train_model_epochs(
            dataloader, loss_fn=loss_fn, spec=spec, evaluator=evaluator
        )

        if evaluator:
            eval_losses = evaluator.restore_models()
            info.update(
                {f"eval_loss(models[{i}])": l for i, l in enumerate(eval_losses)}
            )
        else:
            eval_losses = [np.nan for _ in self.module.models]

        info.update(self.model_grad_info())
        return eval_losses, info

    def _build_dataloader(
        self, train_samples: SampleBatch, loss_fn: Loss, spec: TrainingSpec
    ) -> DataLoader:
        spec = spec.dataloader
        train_tensors = {
            k: self.convert_to_tensor(train_samples[k]) for k in loss_fn.batch_keys
        }
        dataset = TensorDictDataset(train_tensors)
        sampler = RandomSampler(dataset, replacement=spec.replacement)
        return DataLoader(dataset, sampler=sampler, batch_size=spec.batch_size)

    def _setup_evaluator(
        self, eval_samples: Optional[SampleBatch], loss_fn: Loss, spec: TrainingSpec
    ) -> Optional[Evaluator]:
        if not (eval_samples and spec.improvement_threshold is not None):
            return None

        eval_tensors = {
            k: self.convert_to_tensor(eval_samples[k]) for k in loss_fn.batch_keys
        }
        return Evaluator(
            models=self.module.models,
            loss_fn=loss_fn,
            eval_tensors=eval_tensors,
            improvement_threshold=spec.improvement_threshold,
            patience_epochs=spec.patience_epochs,
        )

    def _train_model_epochs(
        self,
        dataloader: DataLoader,
        loss_fn: Loss,
        spec: TrainingSpec,
        evaluator: Optional[Evaluator],
    ) -> StatDict:
        info = {}
        grad_steps = 0
        start = time.time()
        early_stop = False
        epoch = -1
        for epoch in self._model_epochs(spec):
            for minibatch in dataloader:
                with self.optimizers.optimize("models"):
                    losses, train_info = loss_fn(minibatch)
                    losses.sum().backward()

                info.update({"train_" + k: v for k, v in train_info.items()})
                grad_steps += 1
                if spec.max_grad_steps and grad_steps >= spec.max_grad_steps:
                    break

            if evaluator:
                early_stop, eval_info = evaluator.validate(epoch)
                info.update(eval_info)

            if early_stop or self._terminate_epoch(start, grad_steps, spec):
                break

        info["model_epochs"] = epoch + 1
        return info

    @staticmethod
    def _model_epochs(spec: TrainingSpec) -> Iterator[int]:
        return iter(range(spec.max_epochs)) if spec.max_epochs else itertools.count()

    @staticmethod
    def _terminate_epoch(
        start_time: float, model_steps: int, spec: TrainingSpec,
    ) -> bool:
        max_time = spec.max_time or float("inf")
        max_grad_steps = spec.max_grad_steps or float("inf")

        return time.time() - start_time >= max_time or model_steps >= max_grad_steps

    @torch.no_grad()
    def model_grad_info(self) -> StatDict:
        """Returns the average gradient norm accross models."""
        grad_norms = [
            torch.nn.utils.clip_grad_norm_(m.parameters(), float("inf")).item()
            for m in self.module.models
        ]
        return {"grad_norm(models)": np.mean(grad_norms)}

    @staticmethod
    def model_training_defaults():
        """The default configuration dict for model training."""
        return TrainingSpec().to_dict()



# pylint:disable=missing-module-docstring
from ray.rllib.utils import override

from raylab.agents import trainer
from raylab.agents.trainer import Trainer
from raylab.utils.replay_buffer import NumpyReplayBuffer


@trainer.configure
@trainer.option(
    "policy_improvements",
    default=1,
    help="""Policy improvement steps after each sample call to the rollout worker.

    Example:
        With a 'rollout_fragment_length' of 1 and 'policy_improvement' equal to 10,
        will perform 10 policy updates with minibatch size 'train_batch_size' per
        environment step.
    """,
)
@trainer.option(
    "learning_starts",
    default=0,
    help="Sample this many steps before starting optimization.",
)
@trainer.option("train_batch_size", default=128, override=True)
@trainer.option("rollout_fragment_length", default=1, override=True)
@trainer.option("num_workers", default=0, override=True)
@trainer.option("buffer_size", default=500000, help="Size of the replay buffer")
class OffPolicyTrainer(Trainer):
    """Generic trainer for off-policy agents."""

    # pylint:disable=attribute-defined-outside-init
    _name = ""
    _policy = None

    @override(Trainer)
    def _init(self, config, env_creator):
        self.validate_config(config)
        self.workers = self._make_workers(
            env_creator, self._policy, config, num_workers=0
        )
        self.build_replay_buffer(config)

    @override(Trainer)
    def _train(self):
        pre_learning_steps = self.sample_until_learning_starts()
        timesteps_this_iter = 0

        worker = self.workers.local_worker()
        policy = worker.get_policy()
        stats = {}
        while timesteps_this_iter < max(self.config["timesteps_per_iteration"], 1):
            samples = worker.sample()
            timesteps_this_iter += samples.count
            for row in samples.rows():
                self.replay.add(row)
            stats.update(policy.get_exploration_info())

            self._before_replay_steps(policy)
            for _ in range(int(self.config["policy_improvements"])):
                batch = self.replay.sample(self.config["train_batch_size"])
                stats.update(policy.learn_on_batch(batch))
                self.metrics.num_steps_trained += batch.count

        self.metrics.num_steps_sampled += timesteps_this_iter
        return self._log_metrics(stats, timesteps_this_iter + pre_learning_steps)

    def build_replay_buffer(self, config):
        """Construct replay buffer to hold samples."""
        policy = self.get_policy()
        self.replay = NumpyReplayBuffer(
            policy.observation_space, policy.action_space, config["buffer_size"]
        )
        self.replay.seed(config["seed"])

    def sample_until_learning_starts(self):
        """
        Sample enough transtions so that 'learning_starts' steps are collected before
        the next policy update.
        """
        learning_starts = self.config["learning_starts"]
        worker = self.workers.local_worker()
        sample_count = 0
        while self.metrics.num_steps_sampled + sample_count < learning_starts:
            samples = worker.sample()
            sample_count += samples.count
            for row in samples.rows():
                self.replay.add(row)

        if sample_count:
            self.metrics.num_steps_sampled += sample_count
            self.global_vars["timestep"] = self.metrics.num_steps_sampled
            self.workers.foreach_worker(lambda w: w.set_global_vars(self.global_vars))

        return sample_count

    def _before_replay_steps(self, policy):  # pylint:disable=unused-argument
        pass

    def _log_metrics(self, learner_stats, timesteps_this_iter):
        res = self.collect_metrics()
        res.update(timesteps_this_iter=timesteps_this_iter, learner=learner_stats)
        return res

    @staticmethod
    def validate_config(config):
        """Assert configuration values are valid."""
        assert config["num_workers"] == 0, "No point in using additional workers."
        assert (
            config["rollout_fragment_length"] >= 1
        ), "At least one sample must be collected."



"""On-policy Actor-Critic with Normalizing Flows density approximation."""
from raylab.utils.dictionaries import deep_merge

from .abstract import AbstractActorCritic
from .mixins import NormalizingFlowActorMixin
from .mixins import StateValueMixin


BASE_CONFIG = {"actor": {}, "critic": {}}


class OnPolicyNFAC(NormalizingFlowActorMixin, StateValueMixin, AbstractActorCritic):
    """Actor-Critic module with stochastic actor and state-value critics."""

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config, False, ["actor", "critic"])
        config["critic"]["target_vf"] = False
        super().__init__(obs_space, action_space, config)



"""Mixins for Gym environment wrappers."""
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Optional

import gym.utils.seeding as seeding
import numpy as np

from .utils import ignore_rightmost_variables


class RNGMixin:
    """Adds a separate random number generator to an environment wrapper.

    Appends the wrapper's rng seed to the list of seeds returned by
    :meth:`seed`.

    Attributes:
        np_random: A numpy RandomState
    """

    # pylint:disable=missing-function-docstring,too-few-public-methods
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.np_random, _ = seeding.np_random()

    def seed(self, seed: Optional[int] = None) -> List[int]:
        seeds = super().seed(seed) or []
        self.np_random, seed_ = seeding.np_random(seed)
        return seeds + [seed_]


class IrrelevantRedundantMixin(ABC):
    """Common irrelevant/redundant observation wrapper interface."""

    @property
    @abstractmethod
    def added_size(self):
        """Number of right-most irrelevant/redundant state dimensions."""

    def observation(self, observation: np.ndarray) -> np.ndarray:
        """Concatenate irrelevant/redundant variables to the observation."""
        irrelevant_or_redundant = self._added_vars(observation)
        observation = np.concatenate([observation, irrelevant_or_redundant])
        return observation.astype(self.observation_space.dtype)

    @staticmethod
    def wrap_env_function(func: callable, size: int) -> callable:
        """Wrap base env reward/termination function to ignore added variables.

        Args:
            func: Callable for reward/termination function
            size: Number of irrelevant/redundant variables
        """
        return ignore_rightmost_variables(func, size)

    @abstractmethod
    def _added_vars(self, observation: np.ndarray) -> np.ndarray:
        pass

    def _set_reward_if_possible(self):
        if hasattr(self.env, "reward_fn"):
            self.reward_fn = ignore_rightmost_variables(
                self.env.reward_fn, self.added_size
            )

    def _set_termination_if_possible(self):
        if hasattr(self.env, "termination_fn"):
            self.termination_fn = ignore_rightmost_variables(
                self.env.termination_fn, self.added_size
            )
